# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flat

import flatbuffers

class BallPrediction(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsBallPrediction(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = BallPrediction()
        x.Init(buf, n + offset)
        return x

    # BallPrediction
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

# /// A list of places the ball will be at specific times in the future.
# /// It is guaranteed to sorted so that time increases with each slice.
# /// It is NOT guaranteed to have a consistent amount of time between slices.
    # BallPrediction
    def Slices(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from .PredictionSlice import PredictionSlice
            obj = PredictionSlice()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # BallPrediction
    def SlicesLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

def BallPredictionStart(builder): builder.StartObject(1)
def BallPredictionAddSlices(builder, slices): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(slices), 0)
def BallPredictionStartSlicesVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def BallPredictionEnd(builder): return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flat

import flatbuffers

class BoxShape(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsBoxShape(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = BoxShape()
        x.Init(buf, n + offset)
        return x

    # BoxShape
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # BoxShape
    def Length(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

    # BoxShape
    def Width(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

    # BoxShape
    def Height(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

def BoxShapeStart(builder): builder.StartObject(3)
def BoxShapeAddLength(builder, length): builder.PrependFloat32Slot(0, length, 0.0)
def BoxShapeAddWidth(builder, width): builder.PrependFloat32Slot(1, width, 0.0)
def BoxShapeAddHeight(builder, height): builder.PrependFloat32Slot(2, height, 0.0)
def BoxShapeEnd(builder): return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flat

import flatbuffers

class GameInfo(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsGameInfo(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = GameInfo()
        x.Init(buf, n + offset)
        return x

    # GameInfo
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # GameInfo
    def SecondsElapsed(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

    # GameInfo
    def GameTimeRemaining(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

    # GameInfo
    def IsOvertime(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
        return 0

    # GameInfo
    def IsUnlimitedTime(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
        return 0

# /// True when cars are allowed to move, and during the pause menu. False during replays.
    # GameInfo
    def IsRoundActive(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
        return 0

# /// True when the clock is paused due to kickoff, but false during kickoff countdown. In other words, it is true
# /// while cars can move during kickoff. Note that if both players sit still, game clock start and this will become false.
    # GameInfo
    def IsKickoffPause(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
        return 0

# /// Turns true after final replay, the moment the 'winner' screen appears. Remains true during next match
# /// countdown. Turns false again the moment the 'choose team' screen appears.
    # GameInfo
    def IsMatchEnded(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
        return 0

    # GameInfo
    def WorldGravityZ(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

# /// Game speed multiplier, 1.0 is regular game speed.
    # GameInfo
    def GameSpeed(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

# /// Tracks the number of physics frames the game has computed.
# /// May increase by more than one across consecutive packets.
# /// Data type will roll over after 207 days at 120Hz.
    # GameInfo
    def FrameNum(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

def GameInfoStart(builder): builder.StartObject(10)
def GameInfoAddSecondsElapsed(builder, secondsElapsed): builder.PrependFloat32Slot(0, secondsElapsed, 0.0)
def GameInfoAddGameTimeRemaining(builder, gameTimeRemaining): builder.PrependFloat32Slot(1, gameTimeRemaining, 0.0)
def GameInfoAddIsOvertime(builder, isOvertime): builder.PrependBoolSlot(2, isOvertime, 0)
def GameInfoAddIsUnlimitedTime(builder, isUnlimitedTime): builder.PrependBoolSlot(3, isUnlimitedTime, 0)
def GameInfoAddIsRoundActive(builder, isRoundActive): builder.PrependBoolSlot(4, isRoundActive, 0)
def GameInfoAddIsKickoffPause(builder, isKickoffPause): builder.PrependBoolSlot(5, isKickoffPause, 0)
def GameInfoAddIsMatchEnded(builder, isMatchEnded): builder.PrependBoolSlot(6, isMatchEnded, 0)
def GameInfoAddWorldGravityZ(builder, worldGravityZ): builder.PrependFloat32Slot(7, worldGravityZ, 0.0)
def GameInfoAddGameSpeed(builder, gameSpeed): builder.PrependFloat32Slot(8, gameSpeed, 0.0)
def GameInfoAddFrameNum(builder, frameNum): builder.PrependInt32Slot(9, frameNum, 0)
def GameInfoEnd(builder): return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flat

import flatbuffers

# /// Notification that a player triggers some in-game event, such as:
# ///		Win, Loss, TimePlayed;
# ///		Shot, Assist, Center, Clear, PoolShot;
# ///		Goal, AerialGoal, BicycleGoal, BulletGoal, BackwardsGoal, LongGoal, OvertimeGoal, TurtleGoal;
# ///		AerialHit, BicycleHit, BulletHit, /*BackwardsHit,*/ JuggleHit, FirstTouch, BallHit;
# ///		Save, EpicSave, FreezeSave;
# ///		HatTrick, Savior, Playmaker, MVP;
# ///		FastestGoal, SlowestGoal, FurthestGoal, OwnGoal;
# ///		MostBallTouches, FewestBallTouches, MostBoostPickups, FewestBoostPickups, BoostPickups;
# ///		CarTouches, Demolition, Demolish;
# ///		LowFive, HighFive;
class PlayerStatEvent(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsPlayerStatEvent(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = PlayerStatEvent()
        x.Init(buf, n + offset)
        return x

    # PlayerStatEvent
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

# /// index of the player associated with the event
    # PlayerStatEvent
    def PlayerIndex(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

# /// Event type
    # PlayerStatEvent
    def StatType(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return bytes()

def PlayerStatEventStart(builder): builder.StartObject(2)
def PlayerStatEventAddPlayerIndex(builder, playerIndex): builder.PrependInt32Slot(0, playerIndex, 0)
def PlayerStatEventAddStatType(builder, statType): builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(statType), 0)
def PlayerStatEventEnd(builder): return builder.EndObject()



import multiprocessing as mp
import subprocess
from multiprocessing import Event

from rlbot.botmanager.agent_metadata import AgentMetadata
from rlbot.utils.class_importer import import_class_with_base

from rlbot.botmanager.bot_helper_process import BotHelperProcess


class HelperProcessManager:
    """
    Creates and keeps track of BotHelperProcess instances.
    """

    def __init__(self, quit_event: Event):
        self.quit_event = quit_event
        self.helper_process_map = {}

    def start_or_update_helper_process(self, agent_metadata: AgentMetadata):
        """
        Examines the agent metadata to see if the agent needs a helper process. If the process is not running yet,
        create the process. Once the process is running, feed the agent metadata to it.

        If a process is created here, the pid will be added to the agent metadata.
        """

        helper_req = agent_metadata.helper_process_request

        if helper_req is not None:
            if helper_req.key not in self.helper_process_map:
                metadata_queue = mp.Queue()
                if helper_req.python_file_path is not None:
                    process = mp.Process(target=run_helper_process,
                                         args=(helper_req.python_file_path, metadata_queue, self.quit_event,
                                               helper_req.options))
                    process.daemon = True
                    process.start()
                    agent_metadata.pids.add(process.pid)

                    self.helper_process_map[helper_req.key] = metadata_queue

                if helper_req.executable is not None:
                    launch = [helper_req.executable]
                    if helper_req.exe_args is not None:
                        launch.extend(helper_req.exe_args)
                    process = subprocess.Popen(launch, cwd=helper_req.current_working_directory)

                    agent_metadata.pids.add(process.pid)

                    self.helper_process_map[helper_req.key] = metadata_queue

            metadata_queue = self.helper_process_map[helper_req.key]
            metadata_queue.put(agent_metadata)


def run_helper_process(python_file, metadata_queue, quit_event, options):
    """
    :param python_file: The absolute path of a python file containing the helper process that should be run.
    It must define a class which is a subclass of BotHelperProcess.
    :param metadata_queue: A queue from which the helper process will read AgentMetadata updates.
    :param quit_event: An event which should be set when rlbot is shutting down.
    :param options: A dict with arbitrary options that will be passed through to the helper process.
    """
    class_wrapper = import_class_with_base(python_file, BotHelperProcess)
    helper_class = class_wrapper.get_loaded_class()
    helper = helper_class(metadata_queue, quit_event, options)
    helper.start()



import os

from rlbot.agents.base_agent import BOT_CONFIG_AGENT_HEADER
from rlbot.agents.base_independent_agent import BaseIndependentAgent
from rlbot.botmanager import scratch_manager
from rlbot.botmanager.helper_process_request import HelperProcessRequest
from rlbot.parsing.custom_config import ConfigObject


class BaseScratchAgent(BaseIndependentAgent):

    def __init__(self, name, team, index):
        super().__init__(name, team, index)
        self.port: int = None
        self.spawn_browser: bool = False
        self.sb3file: str = None
        self.headless: bool = False
        self.separate_browsers: bool = False
        self.pretend_blue_team: bool = False

    def get_helper_process_request(self):
        file = os.path.realpath(scratch_manager.__file__)
        self.logger.info(self.sb3file)
        port = self.port + self.index if self.separate_browsers else self.port
        key = f'scratch_helper{self.sb3file or ""}-{port}'
        options = {
            'port': port,
            'spawn_browser': self.spawn_browser,
            'sb3-file': self.sb3file,
            'headless': self.headless,
            'pretend_blue_team': self.pretend_blue_team
        }
        return HelperProcessRequest(file, key, options=options)

    def run_independently(self, terminate_request_event):
        pass

    def load_config(self, config_header):
        self.port = config_header.getint('port')
        self.spawn_browser = config_header.getint('spawn_browser')
        self.sb3file = config_header.getpath('sb3file')
        self.headless = config_header.getboolean('headless')
        self.separate_browsers = config_header.getboolean('separate_browsers')
        self.pretend_blue_team = config_header.getboolean('pretend_blue_team')

    @staticmethod
    def create_agent_configurations(config: ConfigObject):
        params = config.get_header(BOT_CONFIG_AGENT_HEADER)
        params.add_value('port', int, default=42008,
                         description='Port to use for websocket communication')
        params.add_value('spawn_browser', bool, default=False,
                         description='True if we should automatically open google chrome to the scratch page.')
        params.add_value('sb3file', str, default=None,
                         description='Location of the scratch .sb3 file to load automatically')
        params.add_value('headless', bool, default=False,
                         description='If true, bot will run automatically with no visible web browser')
        params.add_value('separate_browsers', bool, default=False,
                         description='If true, each bot will get its own separate browser instance '
                                     'and pretend to be player 1. Works best with spawn_browser = True.')
        params.add_value('pretend_blue_team', bool, default=False,
                         description="If true, and if your bots in the browser all on orange, we will pretend in "
                                     "Scratch that they're on the blue team.")



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flat

import flatbuffers

# /// A minimal version of the game tick packet, useful when bandwidth needs to be conserved.
class TinyPacket(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsTinyPacket(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = TinyPacket()
        x.Init(buf, n + offset)
        return x

    # TinyPacket
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # TinyPacket
    def Players(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from .TinyPlayer import TinyPlayer
            obj = TinyPlayer()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # TinyPacket
    def PlayersLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # TinyPacket
    def Ball(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from .TinyBall import TinyBall
            obj = TinyBall()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

def TinyPacketStart(builder): builder.StartObject(2)
def TinyPacketAddPlayers(builder, players): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(players), 0)
def TinyPacketStartPlayersVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def TinyPacketAddBall(builder, ball): builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(ball), 0)
def TinyPacketEnd(builder): return builder.EndObject()



from rlbot.botmanager.helper_process_request import HelperProcessRequest


class AgentMetadata:
    """
    This class is used for storing and passing around the information about a single agent.
    TODO: Currently overlaps with the information that `BotManager` has.
    """

    def __init__(self, index: int, name: str, team: int, pids: set,
                 helper_process_request: HelperProcessRequest = None):
        """
        :param index: The player index, i.e. "this is player number <index>".
        :param name: The agent's name.
        :param team: 0 for blue team or 1 for orange team.
        :param pids: A list all the process ids that this bot uses (including the ones inside the python process).
        :param helper_process_request: A helper process which can be shared, e.g. with other bots of the same type.
        """
        self.index = index
        self.name = name
        self.team = team
        self.pids = pids
        self.helper_process_request = helper_process_request



from pathlib import Path

from rlbot.matchconfig.loadout_config import LoadoutConfig
from rlbot.parsing.agent_config_parser import create_looks_configurations, load_bot_appearance


class BaseLoadoutGenerator:

    def __init__(self, base_directory: Path):
        self.base_directory = base_directory

    def generate_loadout(self, player_index: int, team: int) -> LoadoutConfig:
        raise NotImplementedError("You need to override the generate_loadout function!")

    def load_cfg_file(self, file_path: Path, team: int) -> LoadoutConfig:
        file = self.base_directory / Path(file_path)
        config_object = create_looks_configurations().parse_file(file)
        return load_bot_appearance(config_object, team)



import ctypes


class PlayerInput(ctypes.Structure):
    _fields_ = [("throttle", ctypes.c_float),
                ("steer", ctypes.c_float),
                ("pitch", ctypes.c_float),
                ("yaw", ctypes.c_float),
                ("roll", ctypes.c_float),
                ("jump", ctypes.c_bool),
                ("boost", ctypes.c_bool),
                ("handbrake", ctypes.c_bool),
                ("use_item", ctypes.c_bool)]



class BaseExtension:
    """
    Extend this class to get access to extensions that are run in the game.
    """

    """an instance of SetupManager that contains helpful methods for modifying the game state."""
    setup_manager = None

    def __init__(self, setup_manager):
        self.setup_manager = setup_manager

    def onMatchEnd(self, score, stats):
        """
        Called when a match has ended.
        :param score: This contains a list of scores.  one score per a team.
        :param stats: This contains the stats as how they appear on the scoreboard at the end of the game.
        """
        pass

    def onGoalScored(self, team):
        """
        Called when a goal has been scored.
        :param team: Which team scored the goal.
        """

    def onGoalSaved(self, team):
        """
        Called when a goal has been saved/epic saved
        :param team: The team that saved the ball
        """

    def onMatchStart(self, rlbot_status):
        """
        Called when a match has started
        :return:
        """



import asyncio
import math
import webbrowser
import json
import os
import queue
import time
import shutil
from datetime import datetime, timedelta

import flatbuffers
import websockets
from rlbot.botmanager.agent_metadata import AgentMetadata
from rlbot.botmanager.bot_helper_process import BotHelperProcess
from rlbot.messages.flat import ControllerState, PlayerInput, TinyPacket, TinyPlayer, Vector3, Rotator, \
    TinyBall
from rlbot.utils.logging_utils import get_logger
from rlbot.utils.structures.game_data_struct import GameTickPacket
from rlbot.utils.structures.game_interface import GameInterface
from selenium import webdriver
from selenium.common.exceptions import SessionNotCreatedException
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.wait import WebDriverWait
from webdriver_manager.chrome import ChromeDriverManager

MAX_AGENT_CALL_PERIOD = timedelta(seconds=1.0)


def index_to_player_string(idx):
    return str(idx + 1)


class ScratchManager(BotHelperProcess):

    def __init__(self, agent_metadata_queue, quit_event, options):
        super().__init__(agent_metadata_queue, quit_event, options)
        self.logger = get_logger('scratch_mgr')
        self.game_interface = GameInterface(self.logger)
        self.current_sockets = set()
        self.running_indices = set()
        self.metadata_map = dict()
        self.port: int = options['port']
        self.sb3_file = options['sb3-file']
        self.pretend_blue_team = options['pretend_blue_team']
        self.has_received_input = False
        self.scratch_index_to_rlbot = {}
        self.should_flip_field = False

    async def data_exchange(self, websocket, path):
        async for message in websocket:
            try:
                controller_states = json.loads(message)
                if not self.has_received_input:
                    self.has_received_input = True
                    self.logger.info(f"Just got first input from Scratch {self.sb3_file} {self.port}")

                for key, scratch_state in controller_states.items():
                    scratch_index = int(key)
                    rlbot_index = self.get_rlbot_index(scratch_index)
                    self.game_interface.update_player_input_flat(self.convert_to_flatbuffer(scratch_state, rlbot_index))
            except UnicodeDecodeError:
                backup_location = os.path.join(os.path.dirname(self.sb3_file), f'backup-{str(int(os.path.getmtime(self.sb3_file)))}.sb3')
                print(f"Saving new version of {self.sb3_file} and backing up old version to {backup_location}")
                shutil.move(self.sb3_file, backup_location)
                with open(self.sb3_file, 'wb') as output:
                    output.write(message)

            self.current_sockets.add(websocket)

    def try_receive_agent_metadata(self):
        """
        As agents start up, they will dump their configuration into the metadata_queue.
        Read from it to learn about all the bots intending to use this scratch manager.
        """
        while True:  # will exit on queue.Empty
            try:
                single_agent_metadata: AgentMetadata = self.metadata_queue.get(timeout=0.1)
                self.running_indices.add(single_agent_metadata.index)
                self.metadata_map[single_agent_metadata.index] = single_agent_metadata
            except queue.Empty:
                return
            except Exception as ex:
                self.logger.error(ex)

    def start(self):
        self.logger.info("Starting scratch manager")

        self.game_interface.load_interface()

        # Wait a moment for all agents to have a chance to start up and send metadata
        time.sleep(1)
        self.try_receive_agent_metadata()

        self.logger.info(self.running_indices)

        num_scratch_bots = len(self.running_indices)
        if num_scratch_bots == 0:
            self.logger.error("No scratch bots registered in the scratch manager, exiting!")
            return

        self.setup_index_map()

        all_orange = sum(map(lambda meta: meta.team, self.metadata_map.values())) == len(self.running_indices)
        self.should_flip_field = self.pretend_blue_team and all_orange

        if self.options['spawn_browser']:
            options = Options()
            options.headless = self.options['headless']

            # This prevents an error message about AudioContext when running in headless mode.
            options.add_argument("--autoplay-policy=no-user-gesture-required")

            players_string = ",".join(map(index_to_player_string, range(len(self.running_indices))))

            try:
                driver = webdriver.Chrome(ChromeDriverManager().install(), chrome_options=options)
                driver.get(
                    f"http://scratch.rlbot.org?host=localhost:{str(self.port)}&players={players_string}&awaitBotFile=1")

                if self.sb3_file is not None:
                    element = WebDriverWait(driver, 10).until(
                        EC.presence_of_element_located((By.ID, "sb3-selenium-uploader"))
                    )
                    element.send_keys(self.sb3_file)

            except SessionNotCreatedException:
                # This can happen if the downloaded chromedriver does not match the version of Chrome that is installed.
                webbrowser.open_new(
                    f"http://scratch.rlbot.org?host=localhost:{str(self.port)}&players={players_string}")
                self.logger.info(f"Could not load the Scratch file automatically! You'll need to upload it yourself "
                                 f"from {self.sb3_file}")

        asyncio.get_event_loop().run_until_complete(websockets.serve(self.data_exchange, port=self.port))
        asyncio.get_event_loop().run_until_complete(self.game_loop())

    def setup_index_map(self):
        num_scratch_bots = len(self.running_indices)
        sorted_indices = list(self.running_indices)
        sorted_indices.sort()
        self.scratch_index_to_rlbot = {i: r for i, r in enumerate(sorted_indices)}

        leftovers = [i for i in range(0, max(self.running_indices)) if i not in self.running_indices]
        leftovers.sort()
        for i in range(num_scratch_bots, num_scratch_bots + len(leftovers)):
            self.scratch_index_to_rlbot[i] = leftovers[i - num_scratch_bots]

    def get_rlbot_index(self, scratch_index):
        if scratch_index in self.scratch_index_to_rlbot:
            return self.scratch_index_to_rlbot[scratch_index]
        return scratch_index

    async def game_loop(self):

        last_tick_game_time = None  # What the tick time of the last observed tick was
        last_call_real_time = datetime.now()  # When we last called the Agent

        packet = GameTickPacket()

        # Run until main process tells to stop
        while not self.quit_event.is_set():
            before = datetime.now()

            self.game_interface.update_live_data_packet(packet)

            # Run the Agent only if the gameInfo has updated.
            tick_game_time = packet.game_info.seconds_elapsed
            worth_communicating = tick_game_time != last_tick_game_time or \
                                  datetime.now() - last_call_real_time >= MAX_AGENT_CALL_PERIOD

            ball = packet.game_ball
            if ball is not None and worth_communicating and max(self.running_indices) < packet.num_cars:
                last_tick_game_time = tick_game_time
                last_call_real_time = datetime.now()

                tiny_player_offsets = []
                builder = flatbuffers.Builder(0)

                for i in range(packet.num_cars):
                    tiny_player_offsets.append(self.copy_player(packet.game_cars[i], builder))

                TinyPacket.TinyPacketStartPlayersVector(builder, packet.num_cars)
                for i in reversed(range(0, len(tiny_player_offsets))):
                    rlbot_index = self.get_rlbot_index(i)
                    builder.PrependUOffsetTRelative(tiny_player_offsets[rlbot_index])
                players_offset = builder.EndVector(len(tiny_player_offsets))

                ballOffset = self.copy_ball(ball, builder)

                TinyPacket.TinyPacketStart(builder)
                TinyPacket.TinyPacketAddPlayers(builder, players_offset)
                TinyPacket.TinyPacketAddBall(builder, ballOffset)
                packet_offset = TinyPacket.TinyPacketEnd(builder)

                builder.Finish(packet_offset)
                buffer = bytes(builder.Output())

                filtered_sockets = {s for s in self.current_sockets if s.open}
                for socket in filtered_sockets:
                    await socket.send(buffer)

                self.current_sockets = filtered_sockets

            after = datetime.now()
            duration = (after - before).total_seconds()

            sleep_secs = 1 / 60 - duration
            if sleep_secs > 0:
                await asyncio.sleep(sleep_secs)

    def convert_to_flatbuffer(self, json_state: dict, index: int):
        builder = flatbuffers.Builder(0)

        ControllerState.ControllerStateStart(builder)
        ControllerState.ControllerStateAddSteer(builder, json_state['steer'])
        ControllerState.ControllerStateAddThrottle(builder, json_state['throttle'])
        ControllerState.ControllerStateAddPitch(builder, json_state['pitch'])
        ControllerState.ControllerStateAddYaw(builder, json_state['yaw'])
        ControllerState.ControllerStateAddRoll(builder, json_state['roll'])
        ControllerState.ControllerStateAddJump(builder, json_state['jump'])
        ControllerState.ControllerStateAddBoost(builder, json_state['boost'])
        ControllerState.ControllerStateAddHandbrake(builder, json_state['handbrake'])

        # This may throw a KeyError for anyone using old cached javascript. You should hard-refresh scratch.rlbot.org.
        ControllerState.ControllerStateAddUseItem(builder, json_state['useItem'])
        controller_state = ControllerState.ControllerStateEnd(builder)

        PlayerInput.PlayerInputStart(builder)
        PlayerInput.PlayerInputAddPlayerIndex(builder, index)
        PlayerInput.PlayerInputAddControllerState(builder, controller_state)
        player_input = PlayerInput.PlayerInputEnd(builder)

        builder.Finish(player_input)
        return builder

    def copy_v3(self, v3, builder):
        if self.should_flip_field:
            return Vector3.CreateVector3(builder, -v3.x, -v3.y, v3.z)
        return Vector3.CreateVector3(builder, v3.x, v3.y, v3.z)

    def copy_rot(self, rot, builder):
        yaw = rot.yaw
        if self.should_flip_field:
            yaw = yaw + math.pi if yaw < 0 else yaw - math.pi
        return Rotator.CreateRotator(builder, rot.pitch, yaw, rot.roll)

    def copy_player(self, player, builder):
        TinyPlayer.TinyPlayerStart(builder)
        TinyPlayer.TinyPlayerAddLocation(builder, self.copy_v3(player.physics.location, builder))
        TinyPlayer.TinyPlayerAddVelocity(builder, self.copy_v3(player.physics.velocity, builder))
        TinyPlayer.TinyPlayerAddRotation(builder, self.copy_rot(player.physics.rotation, builder))
        TinyPlayer.TinyPlayerAddTeam(builder, invert_team(player.team) if self.should_flip_field else player.team)
        TinyPlayer.TinyPlayerAddBoost(builder, player.boost)
        return TinyPlayer.TinyPlayerEnd(builder)

    def copy_ball(self, ball, builder):
        phys = ball.physics
        TinyBall.TinyBallStart(builder)
        TinyBall.TinyBallAddLocation(builder, self.copy_v3(phys.location, builder))
        TinyBall.TinyBallAddVelocity(builder, self.copy_v3(phys.velocity, builder))
        return TinyBall.TinyBallEnd(builder)


def invert_team(team):
    return 0 if team == 1 else 1



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flat

class RumbleOption(object):
    No_Rumble = 0
    Default = 1
    Slow = 2
    Civilized = 3
    Destruction_Derby = 4
    Spring_Loaded = 5
    Spikes_Only = 6
    Spike_Rush = 7




import ctypes


class Struct(ctypes.Structure):
    """
    This class exists to add common python functionality to ctypes.Structure.
    This includes:
        Value equality via the `==` operator.
        Showing contents in `repr(struct)`.
    """

    def __eq__(self, other):
        """
        Note: if your Struct contains pointers, pointer equality is used
        as opposed to following the value at the pointer.
        """
        if type(self) != type(other):
            return False
        for field_name, field_type in self._fields_:
            if getattr(self, field_name) != getattr(other, field_name):
                return False
        return True

    def __repr__(self):
        fields_string = ', '.join(
            f'{field_name}: {repr(getattr(self, field_name))}'
            for field_name, field_type in self._fields_
        )
        return f'{self.__class__.__name__}<{fields_string}>'



from multiprocessing import Event, Queue


class BotHelperProcess:
    """
    This is a base class for a process that can be shared by two or more bots. For example, if those bots are using
    a non-python language like Scratch or Java, the bots will likely want to share a single process that interfaces
    with that other language.
    """

    def __init__(self, metadata_queue: Queue, quit_event: Event, options: dict):
        """
        :param metadata_queue: A multiprocessing queue that receives AgentMetadata objects when
        relevant agents are initialized.
        :param quit_event: A multiprocessing event that will be set when rlbot is trying to quit.
        """
        self.metadata_queue = metadata_queue
        self.quit_event = quit_event
        self.options = options

    def start(self):
        raise NotImplementedError



import configparser
import itertools
import os
from pathlib import Path
from typing import List

from requirements.requirement import Requirement

from rlbot.agents.base_agent import BaseAgent, BOT_CONFIG_MODULE_HEADER, BOT_NAME_KEY, LOOKS_CONFIG_KEY, \
    PYTHON_FILE_KEY, LOGO_FILE_KEY, SUPPORTS_EARLY_START_KEY, LOADOUT_GENERATOR_FILE_KEY, SUPPORTS_STANDALONE
from rlbot.agents.base_loadout_generator import BaseLoadoutGenerator
from rlbot.agents.base_script import SCRIPT_FILE_KEY, BaseScript
from rlbot.agents.rlbot_runnable import RLBotRunnable, REQUIREMENTS_FILE_KEY, REQUIRES_TKINTER, USE_VIRTUAL_ENVIRONMENT_KEY
from rlbot.utils.requirements_management import get_missing_packages, get_packages_needing_upgrade
from rlbot.matchconfig.loadout_config import LoadoutConfig
from rlbot.parsing.agent_config_parser import create_looks_configurations, PARTICIPANT_CONFIGURATION_HEADER, \
    PARTICIPANT_CONFIG_KEY, load_bot_appearance, SCRIPT_CONFIGURATION_HEADER, SCRIPT_CONFIG_KEY
from rlbot.parsing.custom_config import ConfigObject
from rlbot.utils.class_importer import import_class_with_base
from rlbot.utils.logging_utils import get_logger

logger = get_logger('rlbot')


class RunnableConfigBundle:
    def __init__(self, config_directory, config_obj: ConfigObject, config_file_name: str = None):
        self.config_directory = config_directory
        self.config_file_name = config_file_name
        self.config_path = os.path.join(self.config_directory, self.config_file_name)
        self.config_obj = config_obj
        self.base_agent_config = RLBotRunnable.base_create_agent_configurations()
        self.base_agent_config.parse_file(self.config_obj, config_directory=config_directory)
        self.name = config_obj.get(BOT_CONFIG_MODULE_HEADER, BOT_NAME_KEY)
        self.supports_early_start = self.base_agent_config.get(BOT_CONFIG_MODULE_HEADER, SUPPORTS_EARLY_START_KEY)
        self.requirements_file = self.get_absolute_path(BOT_CONFIG_MODULE_HEADER, REQUIREMENTS_FILE_KEY)
        self.use_virtual_environment = self.base_agent_config.getboolean(BOT_CONFIG_MODULE_HEADER, USE_VIRTUAL_ENVIRONMENT_KEY)

    def get_logo_file(self):
        # logo.png is a convention we established during the wintertide tournament.
        logo_name = self.base_agent_config.get(BOT_CONFIG_MODULE_HEADER, LOGO_FILE_KEY) or 'logo.png'
        logo_file = os.path.join(self.config_directory, logo_name)
        if os.path.exists(logo_file):
            return os.path.realpath(logo_file)
        return None

    def get_absolute_path(self, header, key):
        path = self.base_agent_config.get(header, key)
        if path is None:
            return None
        if os.path.isabs(path):
            return path
        if self.config_directory is None:
            raise ValueError(f"Can't locate {path} because it's a relative path and we don't know where to look!")
        joined = os.path.join(self.config_directory, path)
        return os.path.realpath(joined)

    def get_missing_python_packages(self) -> List[Requirement]:
        special_reqs = []
        if self.base_agent_config.getboolean(BOT_CONFIG_MODULE_HEADER, REQUIRES_TKINTER):
            special_reqs.append('tkinter')

        if self.use_virtual_environment:
            # Ignore the requirements file for now, because we will be installing it automatically later!
            return get_missing_packages(requirements_file=None, special_reqs=special_reqs)
        else:
            return get_missing_packages(requirements_file=self.requirements_file, special_reqs=special_reqs)

    def get_python_packages_needing_upgrade(self) -> List[Requirement]:
        if self.requirements_file and not self.use_virtual_environment:
            return get_packages_needing_upgrade(requirements_file=self.requirements_file)
        return []


class BotConfigBundle(RunnableConfigBundle):
    def __init__(self, config_directory, config_obj: ConfigObject, config_file_name: str = None):
        super().__init__(config_directory, config_obj, config_file_name)
        self.base_agent_config = BaseAgent.base_create_agent_configurations()
        self.base_agent_config.parse_file(self.config_obj, config_directory=config_directory)
        self.python_file = self.get_absolute_path(BOT_CONFIG_MODULE_HEADER, PYTHON_FILE_KEY)
        self.supports_standalone = self.base_agent_config.getboolean(BOT_CONFIG_MODULE_HEADER, SUPPORTS_STANDALONE)
        self.looks_path = self.get_absolute_path(BOT_CONFIG_MODULE_HEADER, LOOKS_CONFIG_KEY)
        self.loadout_generator_file = self.get_absolute_path(BOT_CONFIG_MODULE_HEADER, LOADOUT_GENERATOR_FILE_KEY)

    def get_looks_config(self) -> ConfigObject:
        """
        Creates a looks config from the config bundle
        :param config_bundle:
        :return:
        """
        if self.looks_path is None:
            raise ValueError("Cannot get looks config because looks_path is None")
        return create_looks_configurations().parse_file(self.looks_path)

    def generate_loadout_config(self, player_index: int, team: int) -> LoadoutConfig:
        if self.loadout_generator_file:
            try:
                generator_wrapper = import_class_with_base(self.loadout_generator_file, BaseLoadoutGenerator)
                generator_class = generator_wrapper.get_loaded_class()
                generator = generator_class(Path(self.config_directory))
                loadout_config = generator.generate_loadout(player_index, team)
                if loadout_config and isinstance(loadout_config, LoadoutConfig):
                    return loadout_config
                else:
                    logger.warn(f"Generated loadout is invalid, will fall back to file. Loadout was: {loadout_config}")
            except Exception as e:
                logger.warn(f"Failed to generate loadout config, will fall back to file. Error was: {e}")

        config_object = self.get_looks_config()
        return load_bot_appearance(config_object, team)


class ScriptConfigBundle(RunnableConfigBundle):
    def __init__(self, config_directory, config_obj: ConfigObject, config_file_name):
        super().__init__(config_directory, config_obj, config_file_name)
        self.base_agent_config = BaseScript.base_create_agent_configurations()
        self.base_agent_config.parse_file(self.config_obj, config_directory=config_directory)
        self.script_file = self.get_absolute_path(BOT_CONFIG_MODULE_HEADER, SCRIPT_FILE_KEY)
        if not self.script_file:
            raise AttributeError(f"Script config {self.config_path} has no script file configured!")


def get_config_obj_and_directory(config_path):
    if not os.path.isfile(config_path):
        raise FileNotFoundError(f"Could not find config file {config_path}!")
    raw_bot_config = configparser.RawConfigParser()
    raw_bot_config.read(config_path, encoding='utf8')
    config_directory = os.path.dirname(os.path.realpath(config_path))
    return raw_bot_config, config_directory


def get_bot_config_bundle(bot_config_path) -> BotConfigBundle:
    raw_bot_config, config_directory = get_config_obj_and_directory(bot_config_path)
    bundle = BotConfigBundle(config_directory, raw_bot_config, os.path.basename(bot_config_path))
    validate_bot_config(bundle)
    return bundle


def get_script_config_bundle(config_path) -> ScriptConfigBundle:
    raw_bot_config, config_directory = get_config_obj_and_directory(config_path)
    bundle = ScriptConfigBundle(config_directory, raw_bot_config, os.path.basename(config_path))
    return bundle


def validate_bot_config(config_bundle) -> None:
    """
    Checks the config bundle to see whether it has all required attributes.
    """
    if not config_bundle.looks_path or not config_bundle.name:
        raise AttributeError(f"Bot config {config_bundle.config_path} is missing a name or a looks_path!")

    # This will raise an exception if the looks config is malformed
    config_bundle.get_looks_config()


def get_bot_config_bundles(num_participants, config: ConfigObject, config_location, config_bundle_overrides):
    """
    Adds all the config files or config objects.
    :param num_participants:
    :param config:
    :param config_location: The location of the rlbot.cfg file
    :param config_bundle_overrides: These are configs that have been loaded from the gui, they get assigned a bot index.
    :return:
    """
    config_bundles = []
    for i in range(num_participants):
        if i in config_bundle_overrides:
            config_bundles.append(config_bundle_overrides[i])
            logger.debug("Config available")
        else:
            bot_config_relative_path = config.get(PARTICIPANT_CONFIGURATION_HEADER, PARTICIPANT_CONFIG_KEY, i)
            bot_config_path = os.path.join(os.path.dirname(config_location), bot_config_relative_path)
            config_bundles.append(get_bot_config_bundle(bot_config_path))
            logger.debug("Reading raw config")

    return config_bundles


def get_script_config_bundles(match_config: ConfigObject, config_location):
    """
    Adds all the script config bundles found in the match config file.
    Attempts to read indexed config keys until it hits an empty one.
    """
    config_bundles = []
    for i in itertools.count():
        script_config_relative_path = match_config.get(SCRIPT_CONFIGURATION_HEADER, SCRIPT_CONFIG_KEY, i)
        if script_config_relative_path is None or script_config_relative_path == 'None':
            break
        script_config_path = os.path.join(os.path.dirname(config_location), script_config_relative_path)
        config_bundles.append(get_script_config_bundle(script_config_path))

    return config_bundles



from rlbot.parsing.agent_config_parser import add_participant_header, add_scripts_header
from rlbot.parsing.match_settings_config_parser import add_mutator_header, add_match_settings_header
from rlbot.parsing.custom_config import ConfigObject
from rlbot.utils.logging_utils import get_logger


TEAM_CONFIGURATION_HEADER = "Team Configuration"
RLBOT_CONFIGURATION_HEADER = 'RLBot Configuration'
EXTENSION_PATH_KEY = 'extension_path'
NETWORKING_ROLE_KEY = 'networking_role'
NETWORK_ADDRESS_KEY = 'network_address'

logger = get_logger('rlbot')


def create_bot_config_layout():
    config_object = ConfigObject()
    rlbot_header = config_object.add_header_name(RLBOT_CONFIGURATION_HEADER)
    rlbot_header.add_value(EXTENSION_PATH_KEY, str, default=None,
                           description='A path to the extension file we want to load')
    rlbot_header.add_value(NETWORKING_ROLE_KEY, str, default='none',
                           description='Defines the behavior when connecting multiple RLBot instances over a network.')
    rlbot_header.add_value(NETWORK_ADDRESS_KEY, str, default='127.0.0.1',
                           description='The IP address to connect to if networking is desired.')

    team_header = config_object.add_header_name(TEAM_CONFIGURATION_HEADER)
    team_header.add_value("Team Blue Color", int, default=0,
                          description="Changes Blue team color, use 0 to use default color")
    team_header.add_value("Team Blue Name", str, default="Blue",
                          description="Changes the Team name to use instead of 'Blue'")
    team_header.add_value("Team Orange Color", int, default=0,
                          description="Changes Blue team color, use 0 to use default color")
    team_header.add_value("Team Orange Name", str, default="Orange",
                          description="Changes the Team name to use instead of 'Orange'")
    add_match_settings_header(config_object)
    add_mutator_header(config_object)
    add_participant_header(config_object)
    add_scripts_header(config_object)
    return config_object




# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flat

import flatbuffers

class ControllerState(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsControllerState(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = ControllerState()
        x.Init(buf, n + offset)
        return x

    # ControllerState
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

# /// -1 for full reverse, 1 for full forward
    # ControllerState
    def Throttle(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

# /// -1 for full left, 1 for full right
    # ControllerState
    def Steer(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

# /// -1 for nose down, 1 for nose up
    # ControllerState
    def Pitch(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

# /// -1 for full left, 1 for full right
    # ControllerState
    def Yaw(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

# /// -1 for roll left, 1 for roll right
    # ControllerState
    def Roll(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

# /// true if you want to press the jump button
    # ControllerState
    def Jump(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
        return 0

# /// true if you want to press the boost button
    # ControllerState
    def Boost(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
        return 0

# /// true if you want to press the handbrake button
    # ControllerState
    def Handbrake(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
        return 0

# /// true if you want to press the 'use item' button, used in rumble etc.
    # ControllerState
    def UseItem(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
        return 0

def ControllerStateStart(builder): builder.StartObject(9)
def ControllerStateAddThrottle(builder, throttle): builder.PrependFloat32Slot(0, throttle, 0.0)
def ControllerStateAddSteer(builder, steer): builder.PrependFloat32Slot(1, steer, 0.0)
def ControllerStateAddPitch(builder, pitch): builder.PrependFloat32Slot(2, pitch, 0.0)
def ControllerStateAddYaw(builder, yaw): builder.PrependFloat32Slot(3, yaw, 0.0)
def ControllerStateAddRoll(builder, roll): builder.PrependFloat32Slot(4, roll, 0.0)
def ControllerStateAddJump(builder, jump): builder.PrependBoolSlot(5, jump, 0)
def ControllerStateAddBoost(builder, boost): builder.PrependBoolSlot(6, boost, 0)
def ControllerStateAddHandbrake(builder, handbrake): builder.PrependBoolSlot(7, handbrake, 0)
def ControllerStateAddUseItem(builder, useItem): builder.PrependBoolSlot(8, useItem, 0)
def ControllerStateEnd(builder): return builder.EndObject()



from typing import List


class HelperProcessRequest:
    """
    This class allows agents to express their need for a helper process that may be shared with other agents.
    """

    def __init__(self, python_file_path: str, key: str, executable: str = None, options: dict = None,
                 exe_args: List[str] = None, current_working_directory = None):
        """
        :param python_file_path: The file that should be loaded and inspected for a subclass of BotHelperProcess.
        :param key: A key used to control the mapping of helper processes to bots. For example, you could set
        something like 'myBotType-team1' in order to get one shared helper process per team.
        :param executable: A path to an executable that should be run as a separate process
        :param options: A dict with arbitrary options that will be passed through to python-based processes.
        :param exe_args: If you are using an executable, we will pass these args to it, if specified.
        :param current_working_directory: If you are using an executable, we will set this as the current working directory.
        """
        self.python_file_path = python_file_path
        self.key = key
        self.executable = executable
        self.options = options
        self.exe_args = exe_args
        self.current_working_directory = current_working_directory



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flat

class GameMap(object):
    DFHStadium = 0
    Mannfield = 1
    ChampionsField = 2
    UrbanCentral = 3
    BeckwithPark = 4
    UtopiaColiseum = 5
    Wasteland = 6
    NeoTokyo = 7
    AquaDome = 8
    StarbaseArc = 9
    Farmstead = 10
    SaltyShores = 11
    DFHStadium_Stormy = 12
    DFHStadium_Day = 13
    Mannfield_Stormy = 14
    Mannfield_Night = 15
    ChampionsField_Day = 16
    BeckwithPark_Stormy = 17
    BeckwithPark_Midnight = 18
    UrbanCentral_Night = 19
    UrbanCentral_Dawn = 20
    UtopiaColiseum_Dusk = 21
    DFHStadium_Snowy = 22
    Mannfield_Snowy = 23
    UtopiaColiseum_Snowy = 24
    Badlands = 25
    Badlands_Night = 26
    TokyoUnderpass = 27
    Arctagon = 28
    Pillars = 29
    Cosmic = 30
    DoubleGoal = 31
    Octagon = 32
    Underpass = 33
    UtopiaRetro = 34
    Hoops_DunkHouse = 35
    DropShot_Core707 = 36
    ThrowbackStadium = 37
    ForbiddenTemple = 38
    RivalsArena = 39
    Farmstead_Night = 40
    SaltyShores_Night = 41
    NeonFields = 42




import flatbuffers
from typing import Union, List

from rlbot.messages.flat import Float, RotatorPartial, Vector3Partial, DesiredPhysics, DesiredGameState, \
    DesiredCarState, DesiredBallState, DesiredBoostState, DesiredGameInfoState, Bool, ConsoleCommand
from rlbot.utils.structures import game_data_struct
from rlbot.messages.flat import GameTickPacket


class Rotator:

    def __init__(self, pitch=None, yaw=None, roll=None):
        self.pitch = pitch
        self.yaw = yaw
        self.roll = roll

    def convert_to_flat(self, builder):
        if self.pitch is None and self.yaw is None and self.roll is None:
            return None  # If no value has to be set this saves some bytes in the builder, will get handled correctly

        RotatorPartial.RotatorPartialStart(builder)
        if self.pitch is not None:
            RotatorPartial.RotatorPartialAddPitch(builder, Float.CreateFloat(builder, self.pitch))
        if self.yaw is not None:
            RotatorPartial.RotatorPartialAddYaw(builder, Float.CreateFloat(builder, self.yaw))
        if self.roll is not None:
            RotatorPartial.RotatorPartialAddRoll(builder, Float.CreateFloat(builder, self.roll))
        return RotatorPartial.RotatorPartialEnd(builder)


class Vector3:

    def __init__(self, x=None, y=None, z=None):
        self.x = x
        self.y = y
        self.z = z

    def convert_to_flat(self, builder):
        if self.x is None and self.y is None and self.z is None:
            return None  # If no value has to be set this saves some bytes in the builder, will get handled correctly

        Vector3Partial.Vector3PartialStart(builder)
        if self.x is not None:
            Vector3Partial.Vector3PartialAddX(builder, Float.CreateFloat(builder, self.x))
        if self.y is not None:
            Vector3Partial.Vector3PartialAddY(builder, Float.CreateFloat(builder, self.y))
        if self.z is not None:
            Vector3Partial.Vector3PartialAddZ(builder, Float.CreateFloat(builder, self.z))
        return Vector3Partial.Vector3PartialEnd(builder)


class Physics:

    def __init__(self, location: Vector3 = None, rotation: Rotator = None, velocity: Vector3 = None,
                 angular_velocity: Vector3 = None):
        self.location = location
        self.rotation = rotation
        self.velocity = velocity
        self.angular_velocity = angular_velocity

    def convert_to_flat(self, builder):
        location_offset = None if self.location is None else self.location.convert_to_flat(builder)
        rotation_offset = None if self.rotation is None else self.rotation.convert_to_flat(builder)
        velocity_offset = None if self.velocity is None else self.velocity.convert_to_flat(builder)
        angular_velocity_offset = None if self.angular_velocity is None else self.angular_velocity.convert_to_flat(
            builder)

        if location_offset is None and rotation_offset is None and velocity_offset is None and \
                angular_velocity_offset is None:
            return None  # If no value has to be set this saves some bytes in the builder, will get handled correctly

        DesiredPhysics.DesiredPhysicsStart(builder)
        if location_offset is not None:
            DesiredPhysics.DesiredPhysicsAddLocation(builder, location_offset)
        if rotation_offset is not None:
            DesiredPhysics.DesiredPhysicsAddRotation(builder, rotation_offset)
        if velocity_offset is not None:
            DesiredPhysics.DesiredPhysicsAddVelocity(builder, velocity_offset)
        if angular_velocity_offset is not None:
            DesiredPhysics.DesiredPhysicsAddAngularVelocity(builder, angular_velocity_offset)
        return DesiredPhysics.DesiredPhysicsEnd(builder)


class BallState:

    def __init__(self, physics: Physics = None):
        self.physics = physics

    def convert_to_flat(self, builder):
        physics_offset = None if self.physics is None else self.physics.convert_to_flat(builder)

        if physics_offset is not None:
            DesiredBallState.DesiredBallStateStart(builder)
            DesiredBallState.DesiredBallStateAddPhysics(builder, physics_offset)
            return DesiredBallState.DesiredBallStateEnd(builder)
        else:
            return None  # This will get handled correctly and is better than taking space while not having to


class CarState:

    def __init__(self, physics: Physics = None, boost_amount: Float = None, jumped: Bool = None,
                 double_jumped: Bool = None):
        self.physics = physics
        self.boost_amount = boost_amount
        self.jumped = jumped
        self.double_jumped = double_jumped

    def convert_to_flat(self, builder):
        """
        In this conversion, we always want to return a valid flatbuffer pointer even if all the
        contents are blank because sometimes we need to put empty car states into the car list
        to make the indices line up.
        """
        physics_offset = None if self.physics is None else self.physics.convert_to_flat(builder)

        DesiredCarState.DesiredCarStateStart(builder)
        if physics_offset is not None:
            DesiredCarState.DesiredCarStateAddPhysics(builder, physics_offset)
        if self.boost_amount is not None:
            DesiredCarState.DesiredCarStateAddBoostAmount(builder, Float.CreateFloat(builder, self.boost_amount))
        if self.jumped is not None:
            DesiredCarState.DesiredCarStateAddJumped(builder, Bool.CreateBool(builder, self.jumped))
        if self.double_jumped is not None:
            DesiredCarState.DesiredCarStateAddDoubleJumped(builder, Bool.CreateBool(builder, self.double_jumped))
        return DesiredCarState.DesiredCarStateEnd(builder)


class BoostState:

    def __init__(self, respawn_time: float = None):
        self.respawn_time = respawn_time

    def convert_to_flat(self, builder):
        """
        In this conversion, we always want to return a valid flatbuffer pointer even if all the
        contents are blank because sometimes we need to put empty boost states into the boost list
        to make the indices line up.
        """

        DesiredBoostState.DesiredBoostStateStart(builder)
        if self.respawn_time is not None:
            DesiredBoostState.DesiredBoostStateAddRespawnTime(builder, Float.CreateFloat(builder, self.respawn_time))
        return DesiredBoostState.DesiredBoostStateEnd(builder)


class GameInfoState:

    def __init__(self, world_gravity_z: float = None, game_speed: float = None, paused: bool = None,
                 end_match: bool = None):

        self.world_gravity_z = world_gravity_z
        self.game_speed = game_speed
        self.paused = paused
        self.end_match = end_match

    def convert_to_flat(self, builder):

        DesiredGameInfoState.DesiredGameInfoStateStart(builder)
        if self.world_gravity_z is not None:
            DesiredGameInfoState.DesiredGameInfoStateAddWorldGravityZ(
                builder, Float.CreateFloat(builder, self.world_gravity_z))
        if self.game_speed is not None:
            DesiredGameInfoState.DesiredGameInfoStateAddGameSpeed(
                builder, Float.CreateFloat(builder, self.game_speed))
        if self.paused is not None:
            DesiredGameInfoState.DesiredGameInfoStateAddPaused(
                builder, Bool.CreateBool(builder, self.paused))
        if self.end_match is not None:
            DesiredGameInfoState.DesiredGameInfoStateAddEndMatch(
                builder, Bool.CreateBool(builder, self.end_match))
        return DesiredGameInfoState.DesiredGameInfoStateEnd(builder)


class GameState:

    def __init__(self, ball: BallState = None, cars=None, boosts=None, game_info: GameInfoState = None,
                 console_commands: List[str]=[]):
        self.ball = ball
        self.cars = cars
        self.boosts = boosts
        self.game_info = game_info
        self.console_commands = console_commands

    def convert_to_flat(self, builder=None) -> DesiredGameState:
        if self.ball is None and self.cars is None and self.boosts is None and self.game_info is None and \
                len(self.console_commands) == 0:
            return None

        if builder is None:
            builder = flatbuffers.Builder(0)

        ball_offset = None if self.ball is None else self.ball.convert_to_flat(builder)

        car_offsets = []
        if self.cars is not None and len(self.cars) > 0:
            max_index = max(self.cars.keys())
            empty_offset = CarState().convert_to_flat(builder)
            for i in range(0, max_index + 1):
                if i in self.cars:
                    car_offsets.append(self.cars[i].convert_to_flat(builder))
                else:
                    car_offsets.append(empty_offset)

        boost_offsets = []
        if self.boosts is not None and len(self.boosts) > 0:
            max_index = max(self.boosts.keys())
            empty_offset = BoostState().convert_to_flat(builder)
            for i in range(0, max_index + 1):
                if i in self.boosts:
                    boost_offsets.append(self.boosts[i].convert_to_flat(builder))
                else:
                    boost_offsets.append(empty_offset)

        game_info_offset = None if self.game_info is None else self.game_info.convert_to_flat(builder)

        car_list_offset = None
        if len(car_offsets) > 0:
            DesiredGameState.DesiredGameStateStartCarStatesVector(builder, len(car_offsets))
            for i in reversed(range(0, len(car_offsets))):
                builder.PrependUOffsetTRelative(car_offsets[i])
            car_list_offset = builder.EndVector(len(car_offsets))

        boost_list_offset = None
        if len(boost_offsets) > 0:
            DesiredGameState.DesiredGameStateStartBoostStatesVector(builder, len(boost_offsets))
            for i in reversed(range(0, len(boost_offsets))):
                builder.PrependUOffsetTRelative(boost_offsets[i])
            boost_list_offset = builder.EndVector(len(boost_offsets))

        console_commands_offsets = []
        for cmd in self.console_commands:
            str_offset = builder.CreateString(cmd)
            ConsoleCommand.ConsoleCommandStart(builder)
            ConsoleCommand.ConsoleCommandAddCommand(builder, str_offset)
            console_commands_offsets.append(ConsoleCommand.ConsoleCommandEnd(builder))

        console_command_list_offset = None
        if self.console_commands is not None and len(self.console_commands) > 0:
            DesiredGameState.DesiredGameStateStartConsoleCommandsVector(builder, len(console_commands_offsets))
            for i in reversed(range(0, len(console_commands_offsets))):
                builder.PrependUOffsetTRelative(console_commands_offsets[i])
            console_command_list_offset = builder.EndVector(len(console_commands_offsets))

        DesiredGameState.DesiredGameStateStart(builder)
        if ball_offset is not None:
            DesiredGameState.DesiredGameStateAddBallState(builder, ball_offset)
        if car_list_offset is not None:
            DesiredGameState.DesiredGameStateAddCarStates(builder, car_list_offset)
        if boost_list_offset is not None:
            DesiredGameState.DesiredGameStateAddBoostStates(builder, boost_list_offset)
        if game_info_offset is not None:
            DesiredGameState.DesiredGameStateAddGameInfoState(builder, game_info_offset)
        if console_command_list_offset is not None:
            DesiredGameState.DesiredGameStateAddConsoleCommands(builder, console_command_list_offset)

        return DesiredGameState.DesiredGameStateEnd(builder)

    @staticmethod
    def create_from_gametickpacket(game_tick_packet: Union[game_data_struct.GameTickPacket, GameTickPacket.GameTickPacket]):
        game_state = GameState()
        if isinstance(game_tick_packet, game_data_struct.GameTickPacket):
            car_states = {}
            for i in range(game_tick_packet.num_cars):
                car = game_tick_packet.game_cars[i]
                car_states[i] = CarState(physics=Physics(location=Vector3(
                    x=car.physics.location.x,
                    y=car.physics.location.y,
                    z=car.physics.location.z
                ), rotation=Rotator(
                    pitch=car.physics.rotation.pitch,
                    yaw=car.physics.rotation.yaw,
                    roll=car.physics.rotation.roll
                ), velocity=Vector3(
                    x=car.physics.velocity.x,
                    y=car.physics.velocity.y,
                    z=car.physics.velocity.z
                ), angular_velocity=Vector3(
                    x=car.physics.angular_velocity.x,
                    y=car.physics.angular_velocity.y,
                    z=car.physics.angular_velocity.z
                )), boost_amount=car.boost, jumped=car.jumped, double_jumped=car.double_jumped)
            game_state.cars = car_states
            ball = game_tick_packet.game_ball
            game_state.ball = BallState(physics=Physics(location=Vector3(
                x=ball.physics.location.x,
                y=ball.physics.location.y,
                z=ball.physics.location.z
            ), rotation=Rotator(
                pitch=ball.physics.rotation.pitch,
                yaw=ball.physics.rotation.yaw,
                roll=ball.physics.rotation.roll
            ), velocity=Vector3(
                x=ball.physics.velocity.x,
                y=ball.physics.velocity.y,
                z=ball.physics.velocity.z
            ), angular_velocity=Vector3(
                x=ball.physics.angular_velocity.x,
                y=ball.physics.angular_velocity.y,
                z=ball.physics.angular_velocity.z
            )))
        elif isinstance(game_tick_packet, GameTickPacket.GameTickPacket):
            car_states = {}
            for i in range(game_tick_packet.PlayersLength()):
                car = game_tick_packet.Players(i)
                car_states[i] = CarState(physics=Physics(location=Vector3(
                    x=car.Physics().Location().X(),
                    y=car.Physics().Location().Y(),
                    z=car.Physics().Location().Z()
                ), rotation=Rotator(
                    pitch=car.Physics().Rotation().Pitch(),
                    yaw=car.Physics().Rotation().Yaw(),
                    roll=car.Physics().Rotation().Roll()
                ), velocity=Vector3(
                    x=car.Physics().Velocity().X(),
                    y=car.Physics().Velocity().Y(),
                    z=car.Physics().Velocity().Z()
                ), angular_velocity=Vector3(
                    x=car.Physics().AngularVelocity().X(),
                    y=car.Physics().AngularVelocity().Y(),
                    z=car.Physics().AngularVelocity().Z()
                )), boost_amount=car.Boost(), jumped=car.Jumped(), double_jumped=car.DoubleJumped())
            game_state.cars = car_states
            ball = game_tick_packet.Ball()
            game_state.ball = BallState(physics=Physics(location=Vector3(
                x=ball.Physics().Location().X(),
                y=ball.Physics().Location().Y(),
                z=ball.Physics().Location().Z()
            ), rotation=Rotator(
                pitch=ball.Physics().Rotation().Pitch(),
                yaw=ball.Physics().Rotation().Yaw(),
                roll=ball.Physics().Rotation().Roll()
            ), velocity=Vector3(
                x=ball.Physics().Velocity().X(),
                y=ball.Physics().Velocity().Y(),
                z=ball.Physics().Velocity().Z()
            ), angular_velocity=Vector3(
                x=ball.Physics().AngularVelocity().X(),
                y=ball.Physics().AngularVelocity().Y(),
                z=ball.Physics().AngularVelocity().Z()
            )))
        return game_state



import os
import atexit
import configparser


TAGAME_CONFIG_LOCATION = os.path.expanduser("~/Documents/My Games/Rocket League/TAGame/Config")


def mergeTASystemSettings():
    """
    Backs up TASystemSettings.ini and creates a new merged TASystemSettings.ini based on TASystemSettings.RLBot.ini.
    Does nothing if no TASystemSettings.RLBot.ini is found.
    """
    TASystemSettings_path = os.path.join(TAGAME_CONFIG_LOCATION, "TASystemSettings.ini")
    TASystemSettings_user_path = os.path.join(TAGAME_CONFIG_LOCATION, "TASystemSettings.user.ini")
    TASystemSettings_RLBot_path = os.path.join(TAGAME_CONFIG_LOCATION, "TASystemSettings.RLBot.ini")

    # only mess with TASystemSettings if TASystemSettings.RLBot.ini exists
    if not os.path.exists(TASystemSettings_RLBot_path):
        return
    # don't mess with TASystemSettings if it has already been messed with.
    if os.path.exists(TASystemSettings_user_path):
        return

    print("Merging TASystemSettings.")

    atexit.register(cleanUpTASystemSettings)

    os.rename(TASystemSettings_path, TASystemSettings_user_path)

    config = configparser.ConfigParser()
    config.read(TASystemSettings_user_path)
    config.read(TASystemSettings_RLBot_path)
    with open(TASystemSettings_path, 'w') as configfile:
        config.write(configfile)

def cleanUpTASystemSettings():
    """
    Reverses the effects of mergeTASystemSettings.
    """
    TASystemSettings_path = os.path.join(TAGAME_CONFIG_LOCATION, "TASystemSettings.ini")
    TASystemSettings_user_path = os.path.join(TAGAME_CONFIG_LOCATION, "TASystemSettings.user.ini")

    if not os.path.exists(TASystemSettings_user_path):
        return

    print("Reverting TASystemSettings.")
    if os.path.exists(TASystemSettings_path):
        os.remove(TASystemSettings_path)

    os.rename(TASystemSettings_user_path, TASystemSettings_path)

    atexit.unregister(cleanUpTASystemSettings)


from typing import List, Optional

from rlbot.matchcomms.common_uses.common_keys import TARGET_PLAYER_INDEX
from rlbot.matchcomms.shared import JSON
from rlbot.agents.base_agent import BaseAgent

"""
This message instructs a certain player to persist some attribute.
"""

MESSAGE_KEY = 'setattrs'
ALL_KEYS_ALLOWED = None


def make_set_attributes_message(target_player_index: int, attrs: JSON) -> JSON:
    assert type(attrs) is dict
    return {
        TARGET_PLAYER_INDEX: target_player_index,
        MESSAGE_KEY: attrs,
    }

def handle_set_attributes_message(msg: JSON, player: BaseAgent, allowed_keys: Optional[List[str]]=ALL_KEYS_ALLOWED) -> bool:
    """
    Sets attributes on the player if the message was applicable for this task.
    If allowed_keys is specified, any key not in this list is silently dropped.
    Returns True iff the message was fully handled by this function.
    """
    attrs = msg.get(MESSAGE_KEY, None)
    if type(attrs) != dict: return False
    if msg.get(TARGET_PLAYER_INDEX, None) != player.index: return False

    # Copy values from attrs to the player.
    if allowed_keys is ALL_KEYS_ALLOWED:
        for k,v in attrs.items():
            setattr(player, k, v)
    else:
        for k in allowed_keys:
            if k in attrs:
                setattr(player, k, attrs[k])
    return True



import time
from timeit import timeit


# A rate limiter that will deduct from inputted datetime difference
class RateLimiter:
    def __init__(self, permits_per_second):
        self.time_per_tick = 1.0 / float(permits_per_second)
        self.then = time.perf_counter()
        self.accumulator = 0.0

    # Assumes acquring 1 permit
    def acquire(self):
        while True:
            now = time.perf_counter()
            self.accumulator += now - self.then
            self.then = now
            if self.accumulator > self.time_per_tick:
                self.accumulator -= self.time_per_tick
                break
            else:
                time.sleep(0.002)


if __name__ == '__main__':

    # Run a rate limiter test
    for fps in (30, 60, 120, 240):

        print("fps :", fps)

        rate_limiter = RateLimiter(fps)

        def test_function():
            time.sleep(0.003)  # simulated bot get_output time
            rate_limiter.acquire()

        n_times = fps
        time_taken = timeit(test_function, number=n_times)

        average_time = time_taken / n_times

        print("Average time :", average_time * 1000, "ms.")
        print("Frequency :", 1 / average_time, "hz.")
        print("Target is", round(1000 / fps, 5), "ms and", fps, "hz.")
        print()



import time

from rlbot.parsing.custom_config import ConfigObject
from rlbot.utils.logging_utils import get_logger

MUTATOR_CONFIGURATION_HEADER = "Mutator Configuration"
MUTATOR_MATCH_LENGTH = "Match Length"
MUTATOR_MAX_SCORE = "Max Score"
MUTATOR_OVERTIME = "Overtime"
MUTATOR_SERIES_LENGTH = "Series Length"
MUTATOR_GAME_SPEED = "Game Speed"
MUTATOR_BALL_MAX_SPEED = "Ball Max Speed"
MUTATOR_BALL_TYPE = "Ball Type"
MUTATOR_BALL_WEIGHT = "Ball Weight"
MUTATOR_BALL_SIZE = "Ball Size"
MUTATOR_BALL_BOUNCINESS = "Ball Bounciness"
MUTATOR_BOOST_AMOUNT = "Boost Amount"
MUTATOR_RUMBLE = "Rumble"
MUTATOR_BOOST_STRENGTH = "Boost Strength"
MUTATOR_GRAVITY = "Gravity"
MUTATOR_DEMOLISH = "Demolish"
MUTATOR_RESPAWN_TIME = "Respawn Time"

MATCH_CONFIGURATION_HEADER = 'Match Configuration'
PARTICIPANT_COUNT_KEY = 'num_participants'
GAME_MODE = 'game_mode'
GAME_MAP = 'game_map'
SKIP_REPLAYS = 'skip_replays'
INSTANT_START = 'start_without_countdown'
EXISTING_MATCH_BEHAVIOR = 'existing_match_behavior'
ENABLE_LOCKSTEP = 'enable_lockstep'
ENABLE_RENDERING = 'enable_rendering'
ENABLE_STATE_SETTING = 'enable_state_setting'
AUTO_SAVE_REPLAY = 'auto_save_replay'

logger = get_logger('config_parser')

game_mode_types = [
    "Soccer",
    "Hoops",
    "Dropshot",
    "Hockey",
    "Rumble",
    "Heatseeker",
]

# This dict needs to say in its current order, because it gets turned into a list
# and then used for index lookup in the GameMap enum in rlbot.fbs.
# If this becomes a superset of the GameMap enum, that's OK.
game_map_dict = {
    "DFHStadium": "Stadium_P",
    "Mannfield": "EuroStadium_P",
    "ChampionsField": "cs_p",
    "UrbanCentral": "TrainStation_P",
    "BeckwithPark": "Park_P",
    "UtopiaColiseum": "UtopiaStadium_P",
    "Wasteland": "wasteland_s_p",
    "NeoTokyo": "NeoTokyo_Standard_P",
    "AquaDome": "Underwater_P",
    "StarbaseArc": "arc_standard_p",
    "Farmstead": "farm_p",
    "SaltyShores": "beach_P",
    "DFHStadium_Stormy": "Stadium_Foggy_P",
    "DFHStadium_Day": "stadium_day_p",
    "Mannfield_Stormy": "EuroStadium_Rainy_P",
    "Mannfield_Night": "EuroStadium_Night_P",
    "ChampionsField_Day": "cs_day_p",
    "BeckwithPark_Stormy": "Park_Rainy_P",
    "BeckwithPark_Midnight": "Park_Night_P",
    "UrbanCentral_Night": "TrainStation_Night_P",
    "UrbanCentral_Dawn": "TrainStation_Dawn_P",
    "UtopiaColiseum_Dusk": "UtopiaStadium_Dusk_P",
    "DFHStadium_Snowy": "Stadium_Winter_P",
    "Mannfield_Snowy": "eurostadium_snownight_p",
    "UtopiaColiseum_Snowy": "UtopiaStadium_Snow_P",
    "Badlands": "Wasteland_P",
    "Badlands_Night": "Wasteland_Night_P",
    "TokyoUnderpass": "NeoTokyo_P",
    "Arctagon": "ARC_P",
    "Pillars": "Labs_CirclePillars_P",
    "Cosmic": "Labs_Cosmic_V4_P",
    "DoubleGoal": "Labs_DoubleGoal_V2_P",
    "Octagon": "Labs_Octagon_02_P",
    "Underpass": "Labs_Underpass_P",
    "UtopiaRetro": "Labs_Utopia_P",
    "Hoops_DunkHouse": "HoopsStadium_P",
    "DropShot_Core707": "ShatterShot_P",
    "ThrowbackStadium": "ThrowbackStadium_P",
    "ForbiddenTemple": "CHN_Stadium_P",
    "RivalsArena": "cs_hw_p",
    "Farmstead_Night": "Farm_Night_P",
    "SaltyShores_Night": "beach_night_p",
    "NeonFields": "music_p",
}

map_types = list(game_map_dict.keys())

match_length_types = [
    "5 Minutes",
    "10 Minutes",
    "20 Minutes",
    "Unlimited"
]

max_score_types = [
    "Unlimited",
    "1 Goal",
    "3 Goals",
    "5 Goals",
]

overtime_mutator_types = [
    "Unlimited",
    "+5 Max, First Score",
    "+5 Max, Random Team"
]

series_length_mutator_types = [
    "Unlimited",
    "3 Games",
    "5 Games",
    "7 Games",
]

game_speed_mutator_types = [
    "Default",
    "Slo-Mo",
    "Time Warp"
]

ball_max_speed_mutator_types = [
    "Default",
    "Slow",
    "Fast",
    "Super Fast"
]

ball_type_mutator_types = [
    "Default",
    "Cube",
    "Puck",
    "Basketball"
]

ball_weight_mutator_types = [
    "Default",
    "Light",
    "Heavy",
    "Super Light"
]

ball_size_mutator_types = [
    "Default",
    "Small",
    "Large",
    "Gigantic"
]

ball_bounciness_mutator_types = [
    "Default",
    "Low",
    "High",
    "Super High"
]

boost_amount_mutator_types = [
    "Default",
    "Unlimited",
    "Recharge (Slow)",
    "Recharge (Fast)",
    "No Boost"
]

rumble_mutator_types = [
    "None",
    "Default",
    "Slow",
    "Civilized",
    "Destruction Derby",
    "Spring Loaded",
    "Spikes Only",
    "Spike Rush"
]

boost_strength_mutator_types = [
    "1x",
    "1.5x",
    "2x",
    "10x"
]

gravity_mutator_types = [
    "Default",
    "Low",
    "High",
    "Super High"
]

demolish_mutator_types = [
    "Default",
    "Disabled",
    "Friendly Fire",
    "On Contact",
    "On Contact (FF)"
]

respawn_time_mutator_types = [
    "3 Seconds",
    "2 Seconds",
    "1 Second",
    "Disable Goal Reset",
]

existing_match_behavior_types = [
    "Restart If Different",
    "Restart",
    "Continue And Spawn",
]


def add_match_settings_header(config_object):
    match_header = config_object.add_header_name(MATCH_CONFIGURATION_HEADER)
    match_header.add_value(PARTICIPANT_COUNT_KEY, int, default=0,
                           description='Number of bots/players which will be spawned.  We support up to max 64.')
    match_header.add_value(GAME_MODE, str, default="Soccer",
                           description="""What game mode the game should load.
                           Accepted values are "Soccer", "Hoops", "Dropshot", "Hockey", "Rumble", "Heatseeker" """)
    match_header.add_value(GAME_MAP, str, default="DFHStadium",
                           description="""Which map the game should load into. Too many to list.""")
    match_header.add_value(SKIP_REPLAYS, bool, default=False,
                           description="""Automatically skip replays after a goal. Also stops match replays from being saved.""")
    match_header.add_value(INSTANT_START, bool, default=False,
                           description="""Skip the kickoff countdown""")
    match_header.add_value(EXISTING_MATCH_BEHAVIOR, str, default="Restart If Different",
                           description="""What should we do if you click run while a match is already in progress?""")
    match_header.add_value(ENABLE_LOCKSTEP, bool, default=False,
                           description="""If True, the framework will wait for outputs from all bots before advancing to the next frame.""")
    match_header.add_value(ENABLE_RENDERING, bool, default=True,
                           description="""If True, bots will be able to draw lines and text in the game.""")
    match_header.add_value(ENABLE_STATE_SETTING, bool, default=True,
                           description="""If True, bots will be able to teleport cars and the ball etc.""")
    match_header.add_value(AUTO_SAVE_REPLAY, bool, default=False,
                           description="""If True, you will be prompted to save a replay at the end of the match.""")


def add_mutator_header(config_object):
    mutator_header = config_object.add_header_name(MUTATOR_CONFIGURATION_HEADER)
    mutator_header.add_value(MUTATOR_MATCH_LENGTH, str, default="5 Minutes")
    mutator_header.add_value(MUTATOR_MAX_SCORE, str, default="Unlimited")
    mutator_header.add_value(MUTATOR_OVERTIME, str, default="Unlimited")
    mutator_header.add_value(MUTATOR_SERIES_LENGTH, str, default="Unlimited")
    mutator_header.add_value(MUTATOR_GAME_SPEED, str, default="Default")
    mutator_header.add_value(MUTATOR_BALL_MAX_SPEED, str, default="Default")
    mutator_header.add_value(MUTATOR_BALL_TYPE, str, default="Default")
    mutator_header.add_value(MUTATOR_BALL_WEIGHT, str, default="Default")
    mutator_header.add_value(MUTATOR_BALL_SIZE, str, default="Default")
    mutator_header.add_value(MUTATOR_BALL_BOUNCINESS, str, default="Default")
    mutator_header.add_value(MUTATOR_BOOST_AMOUNT, str, default="Default")
    mutator_header.add_value(MUTATOR_RUMBLE, str, default="None")
    mutator_header.add_value(MUTATOR_BOOST_STRENGTH, str, default="1x")
    mutator_header.add_value(MUTATOR_GRAVITY, str, default="Default")
    mutator_header.add_value(MUTATOR_DEMOLISH, str, default="Default")
    mutator_header.add_value(MUTATOR_RESPAWN_TIME, str, default="3 Seconds")


def get_num_players(config):
    """
    Returns the number of players specified by the config parser
    """
    return config.getint(MATCH_CONFIGURATION_HEADER, PARTICIPANT_COUNT_KEY)


def parse_mutator_settings(mutator_settings, config: ConfigObject):
    """
    Assigns the mutator settings to the settings object for the dll
    :param mutator_settings:
    :param config:
    """
    mutator_settings.match_length = safe_get_mutator(match_length_types, config, MUTATOR_MATCH_LENGTH)
    mutator_settings.max_score = safe_get_mutator(max_score_types, config, MUTATOR_MAX_SCORE, {'0': 'Unlimited'})
    mutator_settings.overtime = safe_get_mutator(overtime_mutator_types, config, MUTATOR_OVERTIME)
    mutator_settings.series_length = safe_get_mutator(series_length_mutator_types, config, MUTATOR_SERIES_LENGTH)
    mutator_settings.game_speed = safe_get_mutator(game_speed_mutator_types, config, MUTATOR_GAME_SPEED)
    mutator_settings.ball_max_speed = safe_get_mutator(
        ball_max_speed_mutator_types, config, MUTATOR_BALL_MAX_SPEED, {'0': 'Default'})
    mutator_settings.ball_type = safe_get_mutator(ball_type_mutator_types, config, MUTATOR_BALL_TYPE)
    mutator_settings.ball_weight = safe_get_mutator(ball_weight_mutator_types, config, MUTATOR_BALL_WEIGHT)
    mutator_settings.ball_size = safe_get_mutator(
        ball_size_mutator_types, config, MUTATOR_BALL_SIZE, {'1.0': 'Default'})
    mutator_settings.ball_bounciness = safe_get_mutator(
        ball_bounciness_mutator_types, config, MUTATOR_BALL_BOUNCINESS, {'1.0': 'Default'})
    mutator_settings.boost_amount = safe_get_mutator(boost_amount_mutator_types, config, MUTATOR_BOOST_AMOUNT)
    mutator_settings.rumble = safe_get_mutator(rumble_mutator_types, config, MUTATOR_RUMBLE)
    mutator_settings.boost_strength = safe_get_mutator(
        boost_strength_mutator_types, config, MUTATOR_BOOST_STRENGTH, {'Default': '1x', '1.0': '1x'})
    mutator_settings.gravity = safe_get_mutator(gravity_mutator_types, config, MUTATOR_GRAVITY)
    mutator_settings.demolish = safe_get_mutator(demolish_mutator_types, config, MUTATOR_DEMOLISH)
    mutator_settings.respawn_time = safe_get_mutator(respawn_time_mutator_types, config, MUTATOR_RESPAWN_TIME, {
                                                            '3.0': '3 Seconds', '3': '3 Seconds'})


def safe_get_mutator(mutator_options, config, mutator_name, replacement_table={}):

    value = config.get(MUTATOR_CONFIGURATION_HEADER, mutator_name)

    if value in replacement_table:
        logger.warn('**************************************')
        logger.warn(f'The value you\'ve set for {mutator_name} ({value}) is deprecated and will need to be changed to '
                    f'"{replacement_table[value]}" for the next version. Please check your rlbot.cfg!')
        logger.warn('**************************************')
        time.sleep(2.0)
        value = replacement_table[value]

    if value in mutator_options:
        return value
    else:
        logger.warn('**************************************')
        logger.warn(f'The value you\'ve set for {mutator_name} ({value}) is invalid, and will be ignored. '
                    'Please check your rlbot.cfg!')
        logger.warn('**************************************')
        time.sleep(2.0)
        return mutator_options[0]


def parse_match_settings(match_settings, config: ConfigObject):
    """
    Parses the matching settings modifying the match settings object.
    :param match_settings:
    :param config:
    :return:
    """

    match_settings.game_mode = config.get(MATCH_CONFIGURATION_HEADER, GAME_MODE)
    match_settings.game_map = config.get(MATCH_CONFIGURATION_HEADER, GAME_MAP)
    match_settings.skip_replays = config.getboolean(MATCH_CONFIGURATION_HEADER, SKIP_REPLAYS)
    match_settings.instant_start = config.getboolean(MATCH_CONFIGURATION_HEADER, INSTANT_START)
    match_settings.existing_match_behavior = config.get(MATCH_CONFIGURATION_HEADER, EXISTING_MATCH_BEHAVIOR)
    match_settings.enable_lockstep = config.getboolean(MATCH_CONFIGURATION_HEADER, ENABLE_LOCKSTEP)
    match_settings.enable_rendering = config.getboolean(MATCH_CONFIGURATION_HEADER, ENABLE_RENDERING)
    match_settings.enable_state_setting = config.getboolean(MATCH_CONFIGURATION_HEADER, ENABLE_STATE_SETTING)
    match_settings.auto_save_replay = config.getboolean(MATCH_CONFIGURATION_HEADER, AUTO_SAVE_REPLAY)

    parse_mutator_settings(match_settings.mutators, config)



import os
import platform
import socket
import stat
import subprocess
from dataclasses import dataclass
from enum import IntEnum

import psutil

from rlbot.utils.logging_utils import get_logger, DEFAULT_LOGGER
from rlbot.utils.structures.game_interface import get_dll_directory

# Generated randomly by Kipje13, and confirmed to have no conflict with any common programs
# https://www.adminsub.net/tcp-udp-port-finder/23233
IDEAL_RLBOT_PORT = 23233

# This is the port that Rocket League will use by default if we cannot override it.
DEFAULT_RLBOT_PORT = 50000

if platform.system() == 'Windows':
    executable_name = 'RLBot.exe'
elif platform.system() == 'Linux':
    executable_name = 'RLBot'
elif platform.system() == 'Darwin':
    executable_name = 'RLBot_mac'


class NetworkingRole(IntEnum):
    none = 0
    lan_client = 1
    remote_rlbot_server = 2
    remote_rlbot_client = 3


@dataclass
class LaunchOptions:
    networking_role: NetworkingRole = NetworkingRole.none
    remote_address: str = '127.0.0.1'


def launch(launch_options: LaunchOptions = None):
    port = DEFAULT_RLBOT_PORT
    try:
        desired_port = find_usable_port()
        port = desired_port
    except Exception as e:
        print(str(e))

    path = os.path.join(get_dll_directory(), executable_name)
    if not os.access(path, os.F_OK):
        raise FileNotFoundError(f'Unable to find RLBot binary at {path}! Is your antivirus messing you up? Check '
                            'https://github.com/RLBot/RLBot/wiki/Antivirus-Notes.')
    if not os.access(path, os.X_OK):
        os.chmod(path, stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH)
    if not os.access(path, os.X_OK):
        raise PermissionError('Unable to execute RLBot binary due to file permissions! Is your antivirus messing you up? '
                              f'Check https://github.com/RLBot/RLBot/wiki/Antivirus-Notes. The exact path is {path}')

    args = [path, str(port)]
    if launch_options is not None:
        args.append(launch_options.remote_address)
        args.append(str(int(launch_options.networking_role)))
    print(f"Launching RLBot binary with args {args}")
    if platform.system() != 'Windows':
        # Unix only works this way, not sure why. Windows works better with the array, guards against spaces in path.
        args = ' '.join(args)
    return subprocess.Popen(args, shell=True, cwd=get_dll_directory()), port


def find_existing_process():
    logger = get_logger(DEFAULT_LOGGER)
    for proc in psutil.process_iter():
        try:
            if proc.name() == executable_name:
                if len(proc.cmdline()) > 1:
                    port = int(proc.cmdline()[1])
                    return proc, port
                logger.error(f"Failed to find the RLBot port being used in the process args! Guessing "
                             f"{IDEAL_RLBOT_PORT}.")
                return proc, IDEAL_RLBOT_PORT
        except Exception as e:
            logger.error(f"Failed to read the name of a process while hunting for RLBot.exe: {e}")
    return None, None


def find_usable_port():
    for port_to_test in range(IDEAL_RLBOT_PORT, 65535):
        if is_port_accessible(port_to_test):
            return port_to_test
    raise PermissionError('Unable to find a usable port for running RLBot! Is your antivirus messing you up? '
                          'Check https://github.com/RLBot/RLBot/wiki/Antivirus-Notes')


def is_port_accessible(port):
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
        try:
            sock.bind(('127.0.0.1', port))
            return True
        except:
            return False



from contextlib import contextmanager, closing
from dataclasses import dataclass, field
from enum import Enum
from queue import Queue, Empty  # TODO(python 3.7+): use SimpleQueue
from threading import Thread
from typing import Iterator, Dict, Any, Tuple, Set
from urllib.parse import ParseResult as URL
import asyncio
import json
import socket
import time
from logging import Logger

from websockets.server import WebSocketServerProtocol, WebSocketServer
from websockets.exceptions import ConnectionClosed
import websockets

from rlbot.utils.logging_utils import get_logger
from rlbot.matchcomms.shared import MatchcommsPaths, JSON


logger = get_logger('matchcomms')

@dataclass
class MatchcommsServerThread:
    root_url: URL  # How to connect to the server.
    _server: WebSocketServer
    _event_loop: asyncio.AbstractEventLoop
    _thread: Thread

    def close(self, timeout=1):
        self._event_loop.call_soon_threadsafe(self._event_loop.stop)
        self._thread.join(1)
        assert not self._thread.is_alive()

@contextmanager
def keep_in_set(element: Any, s: set):
    """
    Ensures the element is in the set while the context is running
    and removed afterwards.
    """
    s.add(element)
    logger.debug(f'+ client {element.host}:{element.port} (total of {len(s)} clients)')
    try:
        yield
    finally:
        s.remove(element)
        logger.debug(f'- client {element.host}:{element.port} (total of {len(s)} clients)')

@dataclass
class MatchcommsServer:
    users: Set[WebSocketServerProtocol] = field(default_factory=set) # TODO

    async def handle_connection(self, websocket: WebSocketServerProtocol, path: str):
        assert path == MatchcommsPaths.BROADCAST  # TODO consider using other channels
        with keep_in_set(websocket, self.users):
            try:
                async for message in websocket:
                    logger.debug(f'broadcast: {message}')
                    others = [user for user in self.users if user != websocket]
                    if len(others):
                        await asyncio.wait([
                            asyncio.ensure_future(other.send(message))
                            for other in others
                        ])
            except ConnectionClosed:
                return


def find_free_port() -> int:
    # https://stackoverflow.com/a/45690594
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        s.bind(('', 0))
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return s.getsockname()[1]


def launch_matchcomms_server() -> MatchcommsServerThread:
    """
    Launches a background process that handles match communications.
    """
    host = 'localhost'
    port = find_free_port()  # deliberately not using a fixed port to prevent hardcoding fragility.

    event_loop = asyncio.new_event_loop()
    matchcomms_server = MatchcommsServer()
    start_server = websockets.serve(matchcomms_server.handle_connection, host, port, loop=event_loop)
    server = event_loop.run_until_complete(start_server)
    thread = Thread(target=event_loop.run_forever, daemon=True)
    thread.start()
    return MatchcommsServerThread(
        root_url=URL(scheme='ws', netloc=f'{host}:{port}', path='', params='', query='', fragment=''),
        _server=server,
        _event_loop=event_loop,
        _thread=thread,
    )


def self_test():
    server = launch_matchcomms_server()
    from rlbot.matchcomms.client import MatchcommsClient
    com1 = MatchcommsClient(server.root_url)
    com2 = MatchcommsClient(server.root_url)
    com1.outgoing_broadcast.put_nowait({'hi': 'there'})
    try:
        print('<', com2.incoming_broadcast.get(timeout=.2))
    except Empty as e:
        print('Did not get stuff from the queue.')
    com1.close()
    com2.close()
    server.close()


if __name__ == '__main__':
    self_test()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flat

import flatbuffers

class MutatorSettings(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsMutatorSettings(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = MutatorSettings()
        x.Init(buf, n + offset)
        return x

    # MutatorSettings
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # MutatorSettings
    def MatchLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int8Flags, o + self._tab.Pos)
        return 0

    # MutatorSettings
    def MaxScore(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int8Flags, o + self._tab.Pos)
        return 0

    # MutatorSettings
    def OvertimeOption(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int8Flags, o + self._tab.Pos)
        return 0

    # MutatorSettings
    def SeriesLengthOption(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int8Flags, o + self._tab.Pos)
        return 0

    # MutatorSettings
    def GameSpeedOption(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int8Flags, o + self._tab.Pos)
        return 0

    # MutatorSettings
    def BallMaxSpeedOption(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int8Flags, o + self._tab.Pos)
        return 0

    # MutatorSettings
    def BallTypeOption(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int8Flags, o + self._tab.Pos)
        return 0

    # MutatorSettings
    def BallWeightOption(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int8Flags, o + self._tab.Pos)
        return 0

    # MutatorSettings
    def BallSizeOption(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int8Flags, o + self._tab.Pos)
        return 0

    # MutatorSettings
    def BallBouncinessOption(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int8Flags, o + self._tab.Pos)
        return 0

    # MutatorSettings
    def BoostOption(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(24))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int8Flags, o + self._tab.Pos)
        return 0

    # MutatorSettings
    def RumbleOption(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(26))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int8Flags, o + self._tab.Pos)
        return 0

    # MutatorSettings
    def BoostStrengthOption(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(28))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int8Flags, o + self._tab.Pos)
        return 0

    # MutatorSettings
    def GravityOption(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(30))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int8Flags, o + self._tab.Pos)
        return 0

    # MutatorSettings
    def DemolishOption(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(32))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int8Flags, o + self._tab.Pos)
        return 0

    # MutatorSettings
    def RespawnTimeOption(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(34))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int8Flags, o + self._tab.Pos)
        return 0

def MutatorSettingsStart(builder): builder.StartObject(16)
def MutatorSettingsAddMatchLength(builder, matchLength): builder.PrependInt8Slot(0, matchLength, 0)
def MutatorSettingsAddMaxScore(builder, maxScore): builder.PrependInt8Slot(1, maxScore, 0)
def MutatorSettingsAddOvertimeOption(builder, overtimeOption): builder.PrependInt8Slot(2, overtimeOption, 0)
def MutatorSettingsAddSeriesLengthOption(builder, seriesLengthOption): builder.PrependInt8Slot(3, seriesLengthOption, 0)
def MutatorSettingsAddGameSpeedOption(builder, gameSpeedOption): builder.PrependInt8Slot(4, gameSpeedOption, 0)
def MutatorSettingsAddBallMaxSpeedOption(builder, ballMaxSpeedOption): builder.PrependInt8Slot(5, ballMaxSpeedOption, 0)
def MutatorSettingsAddBallTypeOption(builder, ballTypeOption): builder.PrependInt8Slot(6, ballTypeOption, 0)
def MutatorSettingsAddBallWeightOption(builder, ballWeightOption): builder.PrependInt8Slot(7, ballWeightOption, 0)
def MutatorSettingsAddBallSizeOption(builder, ballSizeOption): builder.PrependInt8Slot(8, ballSizeOption, 0)
def MutatorSettingsAddBallBouncinessOption(builder, ballBouncinessOption): builder.PrependInt8Slot(9, ballBouncinessOption, 0)
def MutatorSettingsAddBoostOption(builder, boostOption): builder.PrependInt8Slot(10, boostOption, 0)
def MutatorSettingsAddRumbleOption(builder, rumbleOption): builder.PrependInt8Slot(11, rumbleOption, 0)
def MutatorSettingsAddBoostStrengthOption(builder, boostStrengthOption): builder.PrependInt8Slot(12, boostStrengthOption, 0)
def MutatorSettingsAddGravityOption(builder, gravityOption): builder.PrependInt8Slot(13, gravityOption, 0)
def MutatorSettingsAddDemolishOption(builder, demolishOption): builder.PrependInt8Slot(14, demolishOption, 0)
def MutatorSettingsAddRespawnTimeOption(builder, respawnTimeOption): builder.PrependInt8Slot(15, respawnTimeOption, 0)
def MutatorSettingsEnd(builder): return builder.EndObject()



from typing import Dict

import queue
import time
import traceback

from rlbot.botmanager.agent_metadata import AgentMetadata
from rlbot.botmanager.bot_helper_process import BotHelperProcess
from rlbot.messages.flat import MatchSettings
from rlbot.utils.game_state_util import GameState
from rlbot.utils.logging_utils import get_logger
from rlbot.utils.rendering.rendering_manager import RenderingManager
from rlbot.utils.structures.ball_prediction_struct import BallPrediction
from rlbot.utils.structures.bot_input_struct import PlayerInput
from rlbot.utils.structures.game_data_struct import GameTickPacket, FieldInfoPacket
from rlbot.utils.structures.game_interface import GameInterface


class PythonHivemind(BotHelperProcess):

    # Terminology:
    # hivemind - main process controlling the drones.
    # drone - a bot under the hivemind's control.

    # Path to the executable. NOT USED BY PYTHON HIVEMINDS!
    exec_path = None

    def __init__(self, agent_metadata_queue, quit_event, options):
        super().__init__(agent_metadata_queue, quit_event, options)

        # Sets up the logger. The string is the name of your hivemind.
        # This is configured inside your config file under hivemind_name.
        self.logger = get_logger(options['name'])

        # Give a warning if exec_path was given.
        if self.exec_path is not None:
            self.logger.warning("An exec_path was given, but this is a PythonHivemind, and so it is ignored.")
            self.logger.warning("  Try subclassing SubprocessHivemind if you want to use an executable.")

        # The game interface is how you get access to things
        # like ball prediction, the game tick packet, or rendering.
        self.game_interface = GameInterface(self.logger)

        # drone_indices is a set of bot indices
        # which requested this hivemind with the same key.
        self.drone_indices = set()

        self._field_info = FieldInfoPacket()

    def try_receive_agent_metadata(self):
        """Adds all drones with the correct key to our set of running indices."""
        while not self.metadata_queue.empty():
            # Tries to get the next agent from the queue.
            single_agent_metadata: AgentMetadata = self.metadata_queue.get(timeout=1.0)
            # Adds the index to the drone_indices.
            self.drone_indices.add(single_agent_metadata.index)

    def start(self):
        """Starts the BotHelperProcess - Hivemind."""
        # Prints an activation message into the console.
        # This lets you know that the process is up and running.
        self.logger.debug("Hello, world!")

        # Loads game interface.
        self.game_interface.load_interface()

        # Wait for legitimate field info
        while True:
            self.game_interface.update_field_info_packet(self._field_info)
            if self._field_info.num_goals > 0:
                break
            time.sleep(0.2)

        # Collect drone indices that requested a helper process with our key.
        self.logger.info("Collecting drones; give me a moment.")
        self.try_receive_agent_metadata()
        self.logger.info("Ready to go!")

        # Runs the game loop where the hivemind will spend the rest of its time.
        self.__game_loop()

    def __game_loop(self):
        """
        The bot hivemind will stay in this loop for the whole game.
        This is where the initialize_hive and get_outputs functions are called.
        """

        # Creating ball prediction and field info objects to later update in wrapper methods.
        self._ball_prediction = BallPrediction()

        # Create packet object.
        packet = GameTickPacket()
        # Uses one of the drone indices as a key.
        first_index = next(iter(self.drone_indices))
        self.game_interface.fresh_live_data_packet(packet, 20, first_index)

        # Get a rendering manager for the hivemind.
        self.team = packet.game_cars[first_index].team
        self.renderer = self.game_interface.renderer.get_rendering_manager(bot_index=first_index, bot_team=self.team)

        # Initialization step for your hivemind.
        self.initialize_hive(packet)

        while not self.quit_event.is_set():
            try:
                # Updating the packet.
                self.game_interface.fresh_live_data_packet(packet, 20, first_index)

                # Get outputs from hivemind for each bot.
                # Outputs are expected to be a Dict[int, PlayerInput]
                outputs = self.get_outputs(packet)

                if outputs is None:
                    self.logger.error("No outputs were returned.")
                    self.logger.error("  Try putting `return {i: PlayerInput() for i in self.drone_indices}`")
                    self.logger.error("  in `get_outputs()` to troubleshoot.")
                    continue

                if len(outputs) < len(self.drone_indices):
                    self.logger.error("Not enough outputs were given.")

                elif len(outputs) > len(self.drone_indices):
                    self.logger.error("Too many outputs were given.")

                # Send the drone inputs to the drones.
                for index in outputs:
                    if index not in self.drone_indices:
                        self.logger.error("Tried to send output to bot index not in drone_indices.")
                        continue
                    output = outputs[index]
                    self.game_interface.update_player_input(output, index)

            except Exception:
                traceback.print_exc()


    # Override these methods:

    def initialize_hive(self, packet: GameTickPacket) -> None:
        """
        Override this method if you want to have an initialization step for your hivemind.
        """
        pass

    def get_outputs(self, packet: GameTickPacket) -> Dict[int, PlayerInput]:
        """Where all the logic of your hivemind gets its input and returns its outputs for each drone.

        Use self.drone_indices to access the set of bot indices your hivemind controls.

        Arguments:
            packet {GameTickPacket} -- see https://github.com/RLBot/RLBot/wiki/Input-and-Output-Data-(current)

        Returns:
            Dict[int, PlayerInput] -- A dictionary with drone indices as keys and corresponding PlayerInputs as values.
        """
        return {index: PlayerInput() for index in self.drone_indices}

    # Wrapper methods to make them the same as if you were making a normal python bot:

    def get_ball_prediction_struct(self) -> BallPrediction:
        self.game_interface.update_ball_prediction(self._ball_prediction)
        return self._ball_prediction

    def get_field_info(self) -> FieldInfoPacket:
        # Field info does not need to be updated.
        return self._field_info

    def get_match_settings(self) -> MatchSettings:
        return self.game_interface.get_match_settings()

    def set_game_state(self, game_state: GameState) -> None:
        self.game_interface.set_game_state(game_state)

    # Not everything is supported yet, e.g. quick chats and match comms.



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flat

import flatbuffers

class GameTickPacket(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsGameTickPacket(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = GameTickPacket()
        x.Init(buf, n + offset)
        return x

    # GameTickPacket
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # GameTickPacket
    def Players(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from .PlayerInfo import PlayerInfo
            obj = PlayerInfo()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # GameTickPacket
    def PlayersLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # GameTickPacket
    def BoostPadStates(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from .BoostPadState import BoostPadState
            obj = BoostPadState()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # GameTickPacket
    def BoostPadStatesLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # GameTickPacket
    def Ball(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from .BallInfo import BallInfo
            obj = BallInfo()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # GameTickPacket
    def GameInfo(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from .GameInfo import GameInfo
            obj = GameInfo()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # GameTickPacket
    def TileInformation(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from .DropshotTile import DropshotTile
            obj = DropshotTile()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # GameTickPacket
    def TileInformationLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # GameTickPacket
    def Teams(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from .TeamInfo import TeamInfo
            obj = TeamInfo()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # GameTickPacket
    def TeamsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

def GameTickPacketStart(builder): builder.StartObject(6)
def GameTickPacketAddPlayers(builder, players): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(players), 0)
def GameTickPacketStartPlayersVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def GameTickPacketAddBoostPadStates(builder, boostPadStates): builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(boostPadStates), 0)
def GameTickPacketStartBoostPadStatesVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def GameTickPacketAddBall(builder, ball): builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(ball), 0)
def GameTickPacketAddGameInfo(builder, gameInfo): builder.PrependUOffsetTRelativeSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(gameInfo), 0)
def GameTickPacketAddTileInformation(builder, tileInformation): builder.PrependUOffsetTRelativeSlot(4, flatbuffers.number_types.UOffsetTFlags.py_type(tileInformation), 0)
def GameTickPacketStartTileInformationVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def GameTickPacketAddTeams(builder, teams): builder.PrependUOffsetTRelativeSlot(5, flatbuffers.number_types.UOffsetTFlags.py_type(teams), 0)
def GameTickPacketStartTeamsVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def GameTickPacketEnd(builder): return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flat

import flatbuffers

class Color(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsColor(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Color()
        x.Init(buf, n + offset)
        return x

    # Color
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Color
    def A(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

    # Color
    def R(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

    # Color
    def G(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

    # Color
    def B(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

def ColorStart(builder): builder.StartObject(4)
def ColorAddA(builder, a): builder.PrependUint8Slot(0, a, 0)
def ColorAddR(builder, r): builder.PrependUint8Slot(1, r, 0)
def ColorAddG(builder, g): builder.PrependUint8Slot(2, g, 0)
def ColorAddB(builder, b): builder.PrependUint8Slot(3, b, 0)
def ColorEnd(builder): return builder.EndObject()



import json
from pathlib import Path

from rlbot.matchconfig.loadout_config import LoadoutConfig, LoadoutPaintConfig
from rlbot.matchconfig.match_config import MatchConfig, ExtensionConfig, MutatorConfig, PlayerConfig, ScriptConfig
from rlbot.parsing.agent_config_parser import PARTICIPANT_CONFIGURATION_HEADER, PARTICIPANT_TEAM, PARTICIPANT_TYPE_KEY, load_bot_appearance, PARTICIPANT_BOT_SKILL_KEY
from rlbot.parsing.bot_config_bundle import BotConfigBundle, get_bot_config_bundles, get_script_config_bundles
from rlbot.parsing.custom_config import ConfigObject
from rlbot.parsing.incrementing_integer import IncrementingInteger
from rlbot.parsing.match_settings_config_parser import get_num_players, parse_match_settings
from rlbot.parsing.rlbot_config_parser import create_bot_config_layout, NETWORKING_ROLE_KEY, NETWORK_ADDRESS_KEY
from rlbot.parsing.rlbot_config_parser import RLBOT_CONFIGURATION_HEADER, EXTENSION_PATH_KEY

# ====== rlbot.cfg -> MatchConfig ======
from rlbot.utils.structures.start_match_structures import MAX_PLAYERS


def read_match_config_from_file(match_config_path: Path) -> MatchConfig:
    """
    Parse the rlbot.cfg file on disk into the python datastructure.
    """
    config_obj = create_bot_config_layout()
    config_obj.parse_file(match_config_path, max_index=MAX_PLAYERS)
    return parse_match_config(config_obj, match_config_path, {}, {})


def parse_match_config(config_parser: ConfigObject, config_location, config_bundle_overrides,
                       looks_config_overrides) -> MatchConfig:

    match_config = MatchConfig()
    match_config.mutators = MutatorConfig()

    # Determine number of participants
    num_players = get_num_players(config_parser)

    parse_match_settings(match_config, config_parser)

    # Retrieve bot config files
    config_bundles = get_bot_config_bundles(num_players, config_parser, config_location, config_bundle_overrides)

    match_config.player_configs = []

    human_index_tracker = IncrementingInteger(0)

    # Set configuration values for bots and store name and team
    for i in range(num_players):

        config_bundle = config_bundles[i]

        looks_config_object = None
        if i in looks_config_overrides:
            looks_config_object = looks_config_overrides[i]

        player_config = _load_bot_config(i, config_bundle, looks_config_object, config_parser, human_index_tracker)

        match_config.player_configs.append(player_config)

    extension_path = config_parser.get(RLBOT_CONFIGURATION_HEADER, EXTENSION_PATH_KEY)
    if extension_path and extension_path != 'None':  # The string 'None' ends up in people's config a lot.
        match_config.extension_config = ExtensionConfig()
        match_config.extension_config.python_file_path = extension_path

    match_config.networking_role = config_parser.get(RLBOT_CONFIGURATION_HEADER, NETWORKING_ROLE_KEY)
    match_config.network_address = config_parser.get(RLBOT_CONFIGURATION_HEADER, NETWORK_ADDRESS_KEY)

    match_config.script_configs = [ScriptConfig(bundle.config_path)
                                   for bundle in get_script_config_bundles(config_parser, config_location)]

    return match_config


def get_team(config, index):
    """
    Returns which team the bot is on (represented by an integer)
    """
    return config.getint(PARTICIPANT_CONFIGURATION_HEADER, PARTICIPANT_TEAM, index)


def get_bot_options(bot_type):
    if bot_type == 'human':
        is_bot = False
        is_rlbot = False
    elif bot_type == 'rlbot':
        is_bot = True
        is_rlbot = True
    elif bot_type == 'psyonix':
        is_bot = True
        is_rlbot = False
    elif bot_type == 'party_member_bot':
        # this is a bot running under a human

        is_rlbot = True
        is_bot = False
    else:
        raise ValueError('participant_type value is not "human", "rlbot", "psyonix" or "party_member_bot"')

    return is_bot, is_rlbot


def _load_bot_config(index, config_bundle: BotConfigBundle,
                     looks_config_object: ConfigObject, overall_config: ConfigObject,
                     human_index_tracker: IncrementingInteger) -> PlayerConfig:
    """
    Loads the config data of a single bot
    :param index: This is the bot index (where it appears in game_cars)
    :param bot_configuration: A config object that will eventually be transformed and sent to the game.
    :param config_bundle: A config object for a single bot
    :param overall_config: This is the config for the entire session not one particular bot
    :param human_index_tracker: An object of type HumanIndexManager that helps set human_index correctly.
    :return:
    """

    bot_configuration = PlayerConfig()
    bot_configuration.config_path = config_bundle.config_path

    team_num = get_team(overall_config, index)

    bot_configuration.team = team_num

    # Setting up data about what type of bot it is
    bot_type = overall_config.get(PARTICIPANT_CONFIGURATION_HEADER, PARTICIPANT_TYPE_KEY, index)
    bot_configuration.bot, bot_configuration.rlbot_controlled = get_bot_options(bot_type)
    bot_configuration.bot_skill = overall_config.getfloat(
        PARTICIPANT_CONFIGURATION_HEADER, PARTICIPANT_BOT_SKILL_KEY, index)

    if not bot_configuration.bot:
        bot_configuration.human_index = human_index_tracker.increment()

    # Setting up the bots name
    bot_configuration.name = config_bundle.name

    if looks_config_object:
        loadout_config = load_bot_appearance(looks_config_object, team_num)
    else:
        loadout_config = config_bundle.generate_loadout_config(index, team_num)

    bot_configuration.loadout_config = loadout_config

    return bot_configuration


# ====== MatchConfig -> JSON ======
# The JSON conversion serializes and deserializes MatchConfig in its entirety,
# including player config which is usually specified in different files.

known_types = {
    MatchConfig: '__MatchConfig__',
    MutatorConfig: '__MutatorConfig__',
    ExtensionConfig: '__ExtensionConfig__',
    PlayerConfig: '__PlayerConfig__',
    LoadoutConfig: '__LoadoutConfig__',
    LoadoutPaintConfig: '__LoadoutPaintConfig__',
}


class ConfigJsonEncoder(json.JSONEncoder):
    def default(self, obj):
        for cls, tag in known_types.items():
            if not isinstance(obj, cls):
                continue
            json_obj = obj.__dict__.copy()
            json_obj[tag] = True
            return json_obj
        # Let the base class default method raise the TypeError
        return json.JSONEncoder.default(self, obj)

# ====== JSON -> MatchConfig ======


def as_match_config(json_obj) -> MatchConfig:
    for cls, tag in known_types.items():
        if not json_obj.get(tag, False):
            continue
        obj = cls()
        del json_obj[tag]
        obj.__dict__ = json_obj
        return obj
    return json_obj



# Store the version here so:
# 1) we don't load dependencies by storing it in __init__.py
# 2) we can import it in setup.py for the same reason
# 3) we can import it into your module module
# https://stackoverflow.com/questions/458550/standard-way-to-embed-version-into-python-package

__version__ = '1.55.0'

release_notes = {
    '1.55.0': """
    Using sockets for match start, lifting the restrictions on max players when using flatbuffers,
    and sending a game tick packet immediately upon socket connection to fix some bugs.
    """,
    '1.54.0': """
    Adding a version of the socket management util that uses asyncio.
    """,
    '1.53.3': """
    Making it possible to maintain the map list in the open source portion of RLBot,
    and adding support for NeonFields.
    """,
    '1.52.0': """
    Adding a socket management utility in Python to help people connect directly to
    RLBot.exe and write interesting scripts.
    """,
    '1.51.3': """
    Adding support for standalone python bots, giving faster startup and more
    intuitive development workflow. Also brings support for per-bot virtual environments.
    """,
    '1.50.0': """
    Changing the format of the ready message.
    
    Also:
    - Fixing some bugs related to match settings.
    """,
    '1.49.0': """
    Defaulting the packet send rate to 240 ticks per second, without requiring any .ini file changes.
    See https://github.com/RLBot/RLBot/wiki/Tick-Rate
    """,
    '1.48.6': """
    Using TCP sockets instead of shared memory for transport between RLBot.exe and RLBot_Core_Interface.dll.
    
    Also:
    - Showing ticks per second coming from Rocket League in the performance display.
    """,
    '1.47.0': """
    Trying to fix a bug where low time resolution on Windows causes bots to receive data
    at a lower rate than expected.
    """,
    '1.46.6': """
    Support for locating and launching Rocket League when installed via Epic Games Store,
    thanks to VirxEC. Welcome to all the new people joining with Free to Play!
    """,
    '1.45.0': """
    Fix to support Rocket League patch v1.79.
    """,
    '1.44.6': """
    Splitting the legacy GUI into a separate package to break our dependency on PyQt5.
    If you still really need it, you can find it at https://pypi.org/project/rlbot_legacy_gui/
    
    Also:
    - More elegant handling of Continue and Spawn scenarios.
    """,
    '1.43.7': """
    Bots and scripts can now specify the location of a requirements.txt file which will be analyzed
    by the framework.
    
    Also:
    - Calling begin_rendering and end_rendering on behalf of python bots so they don't need to anymore.
    - Added wait_game_tick_packet helper to the base_script. - L0laapk3
    - Entering 'tkinter' in a requirements.txt file will be understood.
    - Psyonix bots can be preconfigured with a loadout. -az3
    - Clearing car slot in game tick packet when a player leaves.
    - Resetting gravity at the beginning of matches.
    """,
    '1.42.4': """
    System for running multiple scripts that do not own cars in a match.
    
    Also:
    - Set the current working directory of bots to their folder.
    - Do less manipulation of modules when loading bots to reduce weird bugs.
    - Small improvements to hive bot. - Will
    """,
    '1.41.0': """
    Adding more flags for starting matches: enableRendering, enableStateSetting, autoSaveReplay. Also:
    - Mac and Linux users can now use 'q' and 'r' commands to quit or reload bots. - naturecodevoid
    - Update gravity+gamespeed variables in gametickpacket when changed by statesetting. - Kipje13
    """,
    '1.40.1': """
    Support for Heatseeker game mode, including ball prediction thanks to Recruit_main and GodGamer!
    """,
    '1.39.2': """
    You can now generate a fresh loadout for your bot every match using a python script!
    Try setting different colors based on your player index, or a randomized hat.
    Learn how at https://github.com/RLBot/RLBot/wiki/Bot-Customization#loadout-generator

    Also:
    - Fixing color swatch translation. - Darxeal and tarehart
    - Catching errors in hive bots. - Will
    """,
    '1.38.1': """
    Ball prediction is now computed at 120Hz instead of 60Hz!
    
    Note that the data will still only have 60 slices in a second, but the positions will be more
    accurate because of the extra precision behind the scenes. For example, bounces off the goal post
    should be much more reliable now.
    
    Also:
    - Fixed a bug where lockstep could prevent cars from spawning, and freeze forever.
    - Fixed a bug where lockstep freezes forever if Psyonix bots are in the game.
    """,
    '1.37.2': """
    - Better on screen performance measurement for RLBot (hit home key).
    - Bots can now draw when the game is paused, and still move after goal explosions.
    """,
    '1.36.8': """
    Proper appearances for Psyonix bots thanks to Marvin, with an assist from r0bbi3!

    Also:
    - Improved error messages. - Will
    - Fixing bug when trying to spawn tons of cars. - tare
    - Starting on hivemind support classes. Stay tuned! - Will
    - Improving C# documentation. - Redox
    - Feature for loading RLBot-specific game settings (e.g. framerate cap). - L0laapk3
    - Fixed game_tick_packet.game_info.is_match_ended always false - Skyborg
    - Fixed human spawning. - tare
    - Making lockstep work smoother in Python. - tare
    - Lockstep option in old GUI. - Redox
    """,
    '1.35.8': """
    - Adding the Forbidden Temple stadium and a few others.
    - Adding a few custom quick chats.
    - Better support for loading levels that have no goals. - Skyborg
    - Reducing the frequency of python hot reload checks. - Marvin
    - Performance improvement related to logger. - Marvin
    - C# documentation improvements. - Redox
    - Providing a way to get autocomplete on the game tick packet.
    - Trying again to fix inconsistencies with has_wheel_contact.
    - Fixed bug that led to WrongProcessArgs error. - Will
    - Adding a flag to control agent reloads for training. - Marvin
    """,
    '1.34.5': """
    Mac support is here! https://github.com/RLBot/RLBot/wiki/Operating-System-Support

    Also:
    - Improving Remote RLBot so that multiple client computers can connect from behind
      the same IP address.
    - Fixing bug where bots would take a very long time to start if a human was in the
      match config.
    - Fixing input percentages and lockstep feature.
    - Fix inconsistencies with has_wheel_contact. - Kipje13
    """,
    '1.33.1': """
    Remote TCP functionality, primarily for classroom use.
    See https://github.com/RLBot/RLBot/wiki/Remote-RLBot for details!

    Minor updates:
     - Fixing a bug that makes RLBot stutter unpredictably.
    """,
    '1.32.1': """
    - You can specify your car's primary and secondary color swatches by the nearest
      RGB value instead of looking up the color id's on our wiki. You're still constrained
      to the normal color choices, don't get too excited!
      See https://github.com/RLBot/RLBot/wiki/Bot-Customization#colors for details.
    - The lockstep feature should be more reliable now.
    """,
    '1.31.1': """
    Hitbox offsets are now available in the GameTickPacket.
    This should make it possible to use hitboxes without hardcoding any values.

    - Fixes hitboxes issues after RL update.
    """,
    '1.30.2': """
    Kipje13 has a major speedup for rendering in python! You'll get the benefits
    automatically, but ask him for the details if you're curious.

    Also delivering fresher packets to python bots, for improved latency and
    consistency. This may affect bot behavior slightly, check your kickoffs!
    """,
    '1.29.0': """
    Python bots will now be able to specify their preferred tick rate. Do you like running at 120Hz?
    You'll need to add "maximum_tick_rate_preference = 120" to your config file! Example:
    https://github.com/RLBot/RLBot/blob/add379ab700d26f66ee31119a3b4e718a736f7aa/src/test/python/agents/atba/atba.cfg#L9
    The default will be for python bots to run at 60, since that's how most bots have been developed and tested.
    """,
    '1.28.3': """
    Linux Support!

    Kipje13 has really done it, you can run RLBot in Linux now :D
    - Tested in Ubuntu 18.04 with basic python bots, it works great.
    - All the latest features like lockstep are present.
    - This is early days, so many bots and languages won't work yet. Please think of this
      as an alpha and help us find the bugs. Most of the remaining work will be in the
      open source area of RLBot, so drop by #framework-dev if you want to contribute!

    More info at https://github.com/RLBot/RLBot/wiki/Operating-System-Support

    Additional fixes:
    - Open RL using Popen on Linux - cominixo
    - Fixed bug in Scratch manager and switched it to use structs.
    - Ability to save scratch bots over the websocket.
    """,
    '1.27.1': """
    Introducing lockstep mode--if you set enable_lockstep = True in rlbot.cfg, the
    framework will wait for outputs from all bots before advancing to the next frame.

    This feature brought to you by Redox!

    Other features:
    - No longer trying to configure Rocket League with .ini files, which was error-prone.
    """,
    '1.26.4': """
    Adding the ability to execute arbitrary console commands in Rocket League.
    See https://github.com/RLBot/RLBot/wiki/Console-Commands for details!

    Other features:
    - Adding a built-in version of the Scratch bot manager so we can push updates.
    - Adding an early-start system you can opt into if your bot is very slow to start.
      See https://github.com/RLBot/RLBot/wiki/Config-File-Documentation#early-start-system
    - Scratch bots can optionally open a browser-per-bot and pretend to be player 1 on the
      blue team. Good for tournaments!
    - Adding back gravity and game speed setting. - Redox
    """,
    '1.25.5': """
    - Added DummyRenderer, allowing its users to swallow all draw calls. - DomNomNom
    - Fixed a bug where player names with special characters would cause crashes.
    - Giving 'independent' bots more time to shut down.
    - Fixed performance problem with python hot reload for large directories. - RamenAndVitamins
    """,
    '1.24.0': """
    - Car hitbox data is now available for bots.
    - The ball info now includes the shape (sphere, cube, etc.) and size of the ball.
    - Hopefully reduces the amount of crashes when shutting down bot processes.
    """,
    '1.23.2': """
    - Added goal dimensions to field info. - Kipje13
    - Fixed bug where 'restart if different' doesn't start a match once the postgame timer expires.
    - Setting current working directory to the executable location for java and .net bots.
    - Bots can now include a logo! Either put a file called logo.png in the same folder as
      your bot cfg, OR specify logo_file = ... in your bot cfg in the same section as python_file.
      Dimensions of 400x300 are preferred. It will appear soon in RLBotGUI, and maybe on stream!
    """,
    '1.22.2': """
    Fixed the bug that stopped many bots from working after the 1.22.1 update.
    """,
    '1.22.1': """
    Bots can now retrieve the settings object used to launch the current match,
    which can tell you the map, mode, mutator settings, etc.

    Additional fixes:
    - Waiting longer for rlbot to connect.
    - Don't halt forever while waiting for valid packets.
    - Python hot reload watches all your files.
    """,
    '1.21.0': """
    - Latest ball touch now includes a player index value.
    - Fixed blue team colors.
    - Fixed loadout bug that caused incorrect antennas, etc.
    """,
    '1.20.1': """
    Added support for the awesome new game mode Spike Rush!

    To try it, set your mutators:
    Rumble = Spike Rush
    Respawn Time = 1 Second

    To retract your spikes, you have use_item available in your controls output!

    Additional fixes:
    - Will try harder to find an open port, to avoid 'std exception: listen' error.
    - Fixed a bug that prevented multiple .net bots from auto-starting. - Kipje13
    - Fixed a bug that prevented multiple java bots from auto-starting.
    """,
    '1.19.1': """
    Bots can now press the 'use item' button from Rumble mode!
    They don't know what item they have yet...
    """,
    '1.18.1': """
    Created a cleaner option for starting Java bots. Also:
    - seconds_elapsed will pause when the game is paused.
    - game_time_remaining has regained float precision.
    - unlimited_time will be 'true' when using that mutator.
    """,
    '1.17.3': """
    Quickchat is back!

    Kipje13 added quickchat support deep in the framework so bots will be able to
    read it across languages! Everything works exactly the same as before from the
    bot maker's perspective. Stay tuned for instructions on how to read quick
    chat for languages like C# and Java. https://github.com/RLBot/RLBot/wiki/Quickchat

    For now, the chat will use hacky rendering since we don't have official support yet.

    For richer communication with coordinates etc, check out
    https://github.com/RLBot/RLBot/wiki/Matchcomms

    Latest ball touch is also back!
    """,
    '1.16.5': """
    Communication to and between bots is now possible!
    See: https://github.com/RLBot/RLBot/wiki/Matchcomms

    - Support match communication (matchcomms). - DomNomNom
    - Warning when Rocket League is not running in -rlbot mode. - tarehart
    - Fix Python 3.6 compatibility. - DomNomNom
    - Restoring support for rectangle rendering - tarehart
    - The world_gravity_z value will be reported correctly again - tarehart
    - Warnings when you try to use unsupported aspects of state setting - tarehart
    - Avoid needing to manually click Steam confirmation about -rlbot mode. - DomNomNom
    - Brought back the stall mechanic. - tarehart
    """,

    '1.15.10': """
    Read all about our integration with the new Psyonix API!
    https://github.com/RLBot/RLBot/wiki/Psyonix-API-Notes

    Big thanks to Jared Cone and the whole team over at Psyonix!

    New since 1.15.7:
    - Steering will no longer affect dodge angle.
    - You can now specify whether the match should restart via rlbot.cfg.
      This will allow us to experiment with LAN matches.
    """,

    '1.15.7': """
    Integrating with official Psyonix API! Read all about it at
    https://github.com/RLBot/RLBot/wiki/Psyonix-API-Notes

    Big thanks to Jared Cone and the whole team over at Psyonix!

    Post-release fixes:
    - Fixing dodge angle
    - Support for rigid body tick (now including frame counts)
    - Restoring boost pickup timers
    - Rendering turned on by default
    - Possible to launch dropshot / hoops / etc
    - is_bot flag now set correctly
    - Bot performance percentages are back
    - Support for partial values in state setting
    - Fixing the is_super_sonic flag
    - Upgrading to the latest ball prediction code (thanks chip and Kipje13!)
    - Allow a human to play
    - Auto join spectate when there is no human player
    """,

    '1.14.12': """
    - Adding a way of starting matches using a flatbuffer message. - tarehart
    - More accurate get_output call frequency for python bots. - Marvin and chip
    - Fixing compilation of RLBotDotNet project with a breaking change. - tarehart
    - Pinning the psutil package to 5.5.0 to fix 'access denied', 'OSError'. - tarehart
    - Fixed for the friends update. - ccman32
    - Fix for psyonix bots never using boost. - tarehart
    - Avoid killing the parent process, e.g. the GUI, when subprocess agents retire. - tarehart
    - Make the GUI allow non-vital cfg sections to be missing. - tarehart
    - Max 64 bots / players. Be patient while the match is loading :) - skyborg
    - Fix quickchats creating more polling threads each reload. - DomNomNom and SauceTheBoss
    - Fix bot processes not ending when training ends. - DomNomNom
    """,

    '1.13.2': """
    - Fix accidental publishing of pypi package 1.13.1 - DomNomNom
    - Show MIT license in `pip show rlbot`. - DomNomNom
    """,

    '1.12.8': """
    - Support for passing an options dict to BotHelperProcesses. - tarehart
    - Python bots now wait until valid field info to call initialize_agent() - Marvin
    - Field info is no longer being updated each tick and is emptied out if we're not in a game. - Marvin and ccman32
    - Making the details section of bot config files more visible to python GUIs. - tarehart
    - Improved stability of the SetupManager. - DomNomNom
    - Added rendering capability to training exercises. - DomNomNom
    - Allow reading/writing of MatchConfig's including all its depenencies. - DomNomNom
    - Clear the screen when bots retire. - DomNomNom
    - Clear bot inputs when they retire. - DomNomNom
    - Improvements to the controller pass-through agent. - Kipje13 and chip
    - Fix for serialization of multiline values in config files. - tarehart
    """,

    '1.11.1': """
    - Added a new field called 'teams' to packet, which contains goals scored. - Marvin
    - Added team info to latest touch. - Marvin
    - Fix mistake that stopped C# bots from starting. Kipje13
    - GUI now saves relative paths. - Eastvillage
    """,

    '1.10.2': """
    - Bots can now see the match configuration using the new init_match_config method. - tarehart
    - Improved default loadout for bots. - Marvin
    """,

    '1.9.6': """
    - *Much* faster core dll initialization! - ccman32
    - Adding support for a training mode! Check out https://github.com/RLBot/RLBotTraining - DomNomNom
    - Allow the user to change the appearance of human and party-member bot agents via the GUI - r0bbi3
    - Added game speed info to game tick packet and the ability to modify it via state setting - Marvin
    - Make the game stop capturing the mouse cursor if only bots are playing - whatisaphone
    - Various quality-of-life improvements - DomNomNom
    - Match configuration refactoring to make new GUIs easier - tarehart
    """,

    '1.8.3': """
    - Allow SimpleControllerState initialization. - Marvin
    - Passing more params to subprocess agents. - whatisaphone
    - Made game data structs support comparison and repr in python. - DomNomNom
    - Fixing double-logging bug. - Marvin

    For whatisaphone's mouse cursor freedom, roll back to 1.8.2.
    """,

    '1.8.1': """
    - Ability to modify gravity via state setting. Ball prediction reacts properly, and bots are
      informed of the gravity in the game tick packet! - Marvin
    - Sorting the customization items in the GUI. - hallo-doei
    - Making logging more configurable. - Marvin
    - Fixing custom quick chats. - Marvin
    """,

    '1.7.0': """
    The big news: We now support painted items thanks to ccman32!
    See https://github.com/RLBot/RLBot/wiki/Bot-Customization for details.

    Other stuff:
    - Fix for tradeable items not displaying correctly after December update. - ccman32
    - Sending invalid controller inputs will no longer make your bot freeze,
      plus you'll get friendly warnings. - tarehart
    """,

    '1.6.5': """
    Support all subscriptable types for rendering. - Marvin
    """,

    '1.6.4': """
    Fixed compatibility with December update - ccman32
    Added a friendly warning about unsupported python versions - DomNomNom
    Stopped scaring people with statements about locked files - tarehart
    """,

    '1.6.1': """
    Fixed GUI crash when loading certain RLBot config files with relative paths for agents.
    Fixed agent preset loading to allow multiple agents to saved/loaded correctly if they have the same name. - ima9rd
    """,

    '1.6.0': """
    Add support for auto starting .NET executables.
    """,

    '1.5.1': """
    Fixed crash with GUI when no default RLBot.cfg file was found.
    Updated GUI to launch Rocket League when clicking run if no Rocket League process is found. - ima9rd
    """,

    '1.5.0': """
    Adding a have_internet helper function to help streamline upgrade checks. - ima9rd
    """,

    '1.4.2': """
    Adding support for auto-running java bots during tournaments. To take advantage of this
    in your bot, see https://github.com/RLBot/RLBotJavaExample/wiki/Auto-Launching-Java

    Plus bug fixes:
    - Fixed a bug where auto-run executables would crash when trying to write to stderr.
    - Dragging bots to another team in the GUI no longer breaks the config.
    """,

    '1.3.0': """
    Accurate ball prediction for Hoops and Dropshot modes!
      - Kipje13, Marvin, NeverCast, et. al.
    """,

    '1.2.6': """
    Fixed a bug where field info was not extracted properly during dropshot mode.
    It was reporting 2 goals rather than the expected 140.
    """,


    '1.2.5': """
    ***************************************************
    *  Fix for dodge cancels / half flips! - ccman32  *
    ***************************************************

    Plus:
    - Changing the rendering strategy for 3D lines that go past the camera. Formerly it was
      "draw it, even though it's crazy sometimes", now it will be "don't draw it".
    - Showing the rate that inputs are received for each player index when you press the
      [home] key. Toggle back off with the [end] key.
    - Fixed a bug where party_member_bot could get influenced by real controller input.
    - Creating new presets in the GUI works better now.
    - Got rid of the libpng warning seen when using the GUI.
    - Giving specific error messages when cfg files are messed up.
    """,

    '1.2.2': """
    - Rearranged the GUI a bit, and made it load and track appearance configs more effectively.
    - Fixed bug where RUN button behavior in the GUI would not work after killing bots.
    """,

    '1.2.0': """
    - We now offer a 'RigidBodyTick' thanks to whatisaphone! It's a lower-level representation of
    physics data which updates at 120Hz and is not subject to interpolation. You can still make a
    great bot without it, but this feature is quite nice for the scientists among us.

    See https://github.com/RLBot/RLBotPythonExample/wiki/Rigid-Body-Tick for more details!

    - Faster way to access ball prediction data in python. - Skyborg
    """,

    '1.1.3': """
    - Faster way to access ball prediction data in python. - Skyborg
    - Java bots will now shut down when the python framework quits. This has been necessary recently
    to avoid buggy situations.
    - Shutting down the python framework will no longer attempt to kill bots twice in a row.
    - Clicking on the "Run" button twice in a row in the GUI will no longer spawn duplicate processes.
    """,

    '1.1.2': """
    Faster way to access ball prediction data in python. - Skyborg
    """,

    '1.1.1': """
    You can now get information about the ball's status in Dropshot mode thanks to hallo_doei!
    Read all about it at https://github.com/RLBot/RLBot/wiki/Dropshot

    Other changes:
    - The loadout config for orange team is now respected again. - ccman32
    - Fixed a bug where the GUI would crash with a "KeyError". - hallo_doei
    - Avoiding and suppressing some game crashes, and also restoring the
      ability to get game tick data during replays and the postgame. - tarehart
    - Fixed a bug where bots would dodge when they intended to double jump. -tarehart
    """,

    '1.0.6': """
    The latest Rocket League patch broke dodges for our bots; this update fixes it.
    """,

    '1.0.5': """
    Maximum size for a render message has been decreased again because many people experienced
    errors related to memory access. The limit is now only double the original.
    """,

    '1.0.4': """
    - Maximum size for a render message has been increased by a factor of 100. This means you can
      draw a lot of lines at once without getting errors.
    - Boost amount for cars will now round up to the nearest integer, so 0.3% boost will now appear
      as 1 instead of 0.
    - Fixed a crash that would commonly happen after a match ends. As a side effect, you can no longer
      see up-to-date player data during instant replays.
    """,

    '1.0.3': """
    Time for the big 1.0 release! We actually left "beta" a long time ago so this isn't as big
    a milestone as the number implies, but we DO have two great new features!

    1. Setting game state. You can manipulate the position, velocity, etc of the ball and the cars!
    This can be a great help during bot development, and you can also get creative with it. Visit
    the wiki for details and documentation - https://github.com/RLBot/RLBot/wiki/Manipulating-Game-State
    Code written by hallo_doei, ccman32, and tarehart

    2. Ball prediction. We now provide a list of future ball positions based on chip's excellent
    physics modeling. Take advantage of this to do next-level wall reads, catches, and dribbles! You can
    read about the math involved here: https://samuelpmish.github.io/notes/RocketLeague/ball_bouncing/
    Note: currently the wall bounces are only accurate on the standard arena, not hoops or dropshot.
    Documentation and examples can be found here: https://github.com/RLBot/RLBot/wiki/Ball-Path-Prediction
    Code written by chip and tarehart

    Bonus:
     - You can now play on Salty Shores thanks to hallo_doei
     - Bug fix for people with spaces in their file path by Zaptive
     - Subprocess agent for future Rust support by whatisaphone
    """,

    '0.0.32': """
    More comprehensive fix for Rocket League patch 1.50. Compared to previous version:
    - Dropshot tile data is fixed
    - Boost pad data is fixed
    - Loadout configuration is fixed

    Thanks to ccman32 and dtracers for delivering this fix quickly!
    """,

    '0.0.31': """
    Rapid response to Rocket League patch 1.50 with the following known issues:
    - Dropshot tile data is missing
    - Boost pad data is missing
    - Loadout configuration is broken

    Thanks to ccman32 and dtracers for delivering this short-term fix quickly.

    We will follow this up with a proper fix as soon as possible. You may also choose to stay on
    Rocket League 1.49 and RLBot 0.0.30, ask for instructions on discord.
    """,

    '0.0.30': """
    - New core dll that is less likely to break when Rocket League is patched - ccman32 and hallo-doei
    - Fixed bug resulting in incorrect quickchat - dtracers
    - Added more built-in colors to the python rendering manager - Eastvillage
    - Fix for items with a ':' not showing up in the GUI - hallo-doei
    - Fix for GUI not saving correct path - hallo-doei
    - Fix for GUI crash when saving preset then canceling - hallo-doei
    - Adding file checking before injection (Resolves #167) - Redox
    - Fixed typo in rlbot.cfg - Redox
    - Fancy release notes - tarehart and Skyborg
    """
}

release_banner = """

           ______ _     ______       _
     10100 | ___ \ |    | ___ \     | |   00101
    110011 | |_/ / |    | |_/ / ___ | |_  110011
  00110110 |    /| |    | ___ \/ _ \| __| 01101100
    010010 | |\ \| |____| |_/ / (_) | |_  010010
     10010 \_| \_\_____/\____/ \___/ \__| 01001


"""


def get_current_release_notes():
    if __version__ in release_notes:
        return release_notes[__version__]
    return ''


def get_help_text():
    return "Trouble? Ask on Discord at https://discord.gg/5cNbXgG " \
           "or report an issue at https://github.com/RLBot/RLBot/issues"


def print_current_release_notes():
    print(release_banner)
    print(f"Version {__version__}")
    print(get_current_release_notes())
    print(get_help_text())
    print("")



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flat

class GameMode(object):
    Soccer = 0
    Hoops = 1
    Dropshot = 2
    Hockey = 3
    Rumble = 4
    Heatseeker = 5

